import sys


#: Accept all default values without asking user
SLIENTLY_ACCEPT_ALL_DEFAULT_VALUES = False


def prompt(text: str, *args, new_line: bool = True, **kwargs):
    """
    Prompt in terminal (stdout).

    Args:
        text (str): Text.
        *args: More text.
        new_line (bool, optional): End with a new line. Defaults to True.
        **kwargs: Other key word arguments used by :py:meth:`print`.
    """
    line_end = '\n' if new_line else ''
    print(text, *args, file=sys.stdout, end=line_end, **kwargs)
    sys.stdout.flush()


def select(text: str, cases: list, default: int = None, case_sensitive: bool = False):
    """
    Let user select one of the cases.

    Args:
        text (str): Text.
        cases (list[tuple]): Cases, should be list of tuples. Each tuple is in form `('display text', 'user's input')`.
                        For example, `[('(Y)es, 'y'), ('(N)o', 'n')]`.
        default (int, optional): Default case index in `cases`. Empty or space is treated as default case.
                            None means no default case. Defaults to None.
        case_sensitive (bool, optional): If user's input is case sensitive, defaults to False.

    Returns:
        str: User's input.
    """
    prompt(text)
    case_text = []
    for idx, c in enumerate(cases):
        if default is not None and idx == default:
            case_text.append('[{}]'.format(c[0]))
        else:
            case_text.append('{}'.format(c[0]))
    prompt(' / '.join(case_text))
    valid_cases = [c[1] for c in cases]
    if default is not None:
        valid_cases.append('')
    if not case_sensitive:
        valid_cases = list(map(lambda x: x.lower(), valid_cases))

    while True:
        user_input = ''
        if not SLIENTLY_ACCEPT_ALL_DEFAULT_VALUES or default is None:
            user_input = input().strip()
            if not case_sensitive:
                user_input = user_input.lower()
            if user_input not in valid_cases:
                prompt('Invalid input, please retry')
                continue

        if user_input == '' and default is not None:
            return cases[default][1]
        return user_input


def confirm(text: str, default: bool = None):
    """
    Let user choose Yes or No.

    Args:
        text (str): Text.
        default (bool, optional): True sets Yes as default case, False sets No. None means no default case.
                                Defaults to None.

    Returns:
        bool: True means Yes, False means No.
    """
    if default is not None:
        default = 0 if default else 1
    return select(text, cases=[('(Y)es', 'y',), ('(N)o', 'n')], default=default, case_sensitive=False) == 'y'


class Progress(object):
    """
    Progress status.

    Args:
        format_ (str, optional): Format of text.
        start (str, optional): Text while starting.
        end (str, optional): Text while ending.

    Note:

        Please use in `with` statement::

            with rltk.cli.progress(format_='{}%') as p:
                for i in range(11):
                    time.sleep(0.5)
                    p.update(i * 10)

    """

    def __init__(self, format_: str = '{}', start: str = 'Starting...', end: str = 'Done!'):
        self._format = format_
        self._prev_len = 0
        self._start = start
        self._end = end

    def update(self, *args):
        """
        Update progress.

        Args:
            *args: Arguments which will be formatted by `format_`.
        """
        text = self._format.format(*args)

        # clean up
        prompt('\r' + ' ' * self._prev_len, new_line=False)

        # overwrite
        prompt('\r' + text, new_line=False)
        self._prev_len = len(text)

    def __enter__(self):
        """
        Start prompt.
        """
        if self._start:
            prompt(self._start, new_line=False)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        End prompt.
        """
        # clean up
        prompt('\r' + ' ' * self._prev_len, new_line=False)

        if self._end:
            prompt('\r' + self._end, new_line=False)

        # new line
        prompt('')


progress = Progress


def input_(text: str, default: str = None, type_: type = str):
    """
    Input.

    Args:
        text (str): Text.
        default (str, optional): Default value. Defaults to None which means no default value.
        type_ (type, optional): Type of input value, defaults to `str`.

    Returns:
        object: User input in type `type_`.

    Note:
        Make sure default value can be converted by `type_`, otherwise exception will be raised.
    """
    prompt(text)

    while True:
        if not SLIENTLY_ACCEPT_ALL_DEFAULT_VALUES or default is None:
            user_input = input().strip()
            try:
                return type_(user_input)
            except:
                prompt('Invalid input, please retry')
        else:
            return type_(default)




import rltk.utils as utils


def get_ngrams(s, n):
    all_ngrams = list()
    for i in range(len(s) - 1):
        if i + n <= len(s):
            all_ngrams.append(s[i:i + n])

    return set(all_ngrams)


def qgram_distance(s0, s1, n=2):
    """
    QGram Distance is the number of distinct q-grams (n-grams) between 2 strings

    Args:
        s1 (str): Sequence 1.
        s2 (str): Sequence 2.

    Returns:
        float: QGram Distance.

    Examples:
        >>> rltk.qgram_distance('abcde','abdcde')
        3
    """
    if n > max(len(s0), len(s1)):
        return 1

    s0_ngrams = get_ngrams(s0, n)
    s1_ngrams = get_ngrams(s1, n)
    all_ngrams = list(s0_ngrams | s1_ngrams)

    v0 = [1 if all_ngrams[i] in s0 else 0 for i in range(len(all_ngrams))]
    v1 = [1 if all_ngrams[i] in s1 else 0 for i in range(len(all_ngrams))]

    return sum([1 if v0[i] != v1[i] else 0 for i in range(len(v0))])


def qgram_similarity(s0, s1, n=2):
    """
    QGram Similarity is the number of common q-grams (n-grams) between 2 strings

    Args:
        s1 (str): Sequence 1.
        s2 (str): Sequence 2.

    Returns:
        float: QGram Similarity.

    Examples:
        >>> rltk.qgram_similarity('abcde','abdcde')
        3
    """

    if n > max(len(s0), len(s1)):
        return 0

    s0_ngrams = get_ngrams(s0, n)
    s1_ngrams = get_ngrams(s1, n)
    all_ngrams = list(s0_ngrams | s1_ngrams)

    v0 = [1 if all_ngrams[i] in s0 else 0 for i in range(len(all_ngrams))]
    v1 = [1 if all_ngrams[i] in s1 else 0 for i in range(len(all_ngrams))]

    return sum([1 if v0[i] == v1[i] else 0 for i in range(len(v0))])



import rltk.utils as utils


def soundex(s):
    """
    The standard used for this implementation is provided by `U.S. Census Bureau <https://www.archives.gov/research/census/soundex.html>`_.

    Args:
        s (str): Sequence.

    Returns:
        str: Coded sequence.

    Examples:
        >>> rltk.soundex('ashcraft')
        'A261'
        >>> rltk.soundex('pineapple')
        'P514'
    """

    utils.check_for_none(s)
    utils.check_for_type(str, s)

    s = utils.unicode_normalize(s)

    if len(s) == 0:
        raise ValueError('Empty string')

    s = s.upper()

    CODES = (
        ('BFPV', '1'),
        ('CGJKQSXZ', '2'),
        ('DT', '3'),
        ('L', '4'),
        ('MN', '5'),
        ('R', '6'),
        ('AEIOUHWY', '.')  # placeholder
    )
    CODE_DICT = dict((c, replace) for chars, replace in CODES for c in chars)

    sdx = s[0]
    for i in range(1, len(s)):
        if s[i] not in CODE_DICT:
            continue

        code = CODE_DICT[s[i]]
        if code == '.':
            continue
        if s[i] == s[i - 1]:  # ignore same letter
            continue
        if s[i - 1] in CODE_DICT and CODE_DICT[s[i - 1]] == code:  # 'side-by-side' rule
            continue
        if s[i - 1] in ('H', 'W') and i - 2 > 0 and \
                        s[i - 2] in CODE_DICT and CODE_DICT[s[i - 2]] != '.':  # consonant separators
            continue

        sdx += code

    sdx = sdx[0:4].ljust(4, '0')

    return sdx


def soundex_similarity(s1, s2):
    """
    soundex(s1) == soundex(s2)
    
    Args:
        s1 (str): Sequence.
        s2 (str): Sequence.

    Returns:
        float: if soundex(s1) equals to soundex(s2)
    """
    return 1 if soundex(s1) == soundex(s2) else 0



import unicodedata

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from rltk.dataset import Dataset
    from rltk.blocking.block import Block
    from rltk.evaluation.ground_truth import GroundTruth


def check_for_none(*args):
    for arg in args:
        if arg is None:
            raise ValueError('Missing parameter')


def check_for_type(type, *args):
    for arg in args:
        if not isinstance(arg, type):
            raise TypeError('Wrong type of parameter')


def unicode_normalize(s):
    return unicodedata.normalize('NFKD', s)


def convert_list_to_set(s):
    if isinstance(s, list):
        s = set(s)
    return s


def get_record_pairs(dataset1,
                     dataset2,
                     block: 'Block' = None,
                     ground_truth: 'GroundTruth' = None):
    """
    Generate pairs to compare.

    Args:
        dataset1 (Dataset): dataset 1
        dataset2 (Dataset): dataset 2
        block (Block, optional): block 
        ground_truth (GroundTruth, optional): ground truth
    """
    if block and not ground_truth:
        for _, id1, id2 in block.pairwise(dataset1.id, dataset2.id):
            yield dataset1.get_record(id1), dataset2.get_record(id2)
    elif ground_truth and not block:
        for id1, id2, label in ground_truth:
            yield dataset1.get_record(id1), dataset2.get_record(id2)
    elif ground_truth and block:
        for _, id1, id2 in block.pairwise(dataset1.id, dataset2.id):
            if ground_truth.is_member(id1, id2):
                yield dataset1.get_record(id1), dataset2.get_record(id2)
    else:
        for r1 in dataset1:
            for r2 in dataset2:
                yield r1, r2



import math
import rltk.utils as utils


def _jaro_winkler(s1, s2, threshold=0.7, scaling_factor=0.1, prefix_len=4):
    jaro = _jaro_distance(s1, s2)
    if jaro > threshold:
        l = min(len(_get_prefix(s1, s2)), prefix_len)  # max len of common prefix is 4
        jaro += (scaling_factor * l * (1.0 - jaro))
    return jaro


def jaro_winkler_similarity(s1, s2, threshold=0.7, scaling_factor=0.1, prefix_len=4):
    """
    The max length for common prefix is 4.

    Args:
        s1 (str): Sequence 1.
        s2 (str): Sequence 2.
        threshold (int, optional): Boost threshold, prefix bonus is only added when compared strings have a Jaro Distance above it. Defaults to 0.7.
        scaling_factor (int, optional): Scaling factor for how much the score is adjusted upwards for having common prefixes. Defaults to 0.1.

    Returns:
        float: Jaro Winkler Similarity.

    Examples:
        >>> rltk.jaro_winkler_similarity('abchello', 'abcworld')
        0.6833333333333332
        >>> rltk.jaro_winkler_similarity('hello', 'world')
        0.4666666666666666
    """
    return _jaro_winkler(s1, s2, threshold, scaling_factor, prefix_len)


def jaro_winkler_distance(s1, s2, threshold=0.7, scaling_factor=0.1, prefix_len=4):
    """
    Jaro Winkler Distance is computed as 1 - jaro_winkler_similarity.

    Args:
        s1 (str): Sequence 1.
        s2 (str): Sequence 2.
        threshold (int, optional): Boost threshold, prefix bonus is only added when compared strings have a Jaro Distance above it. Defaults to 0.7.
        scaling_factor (int, optional): Scaling factor for how much the score is adjusted upwards for having common prefixes. Defaults to 0.1.

    Returns:
        float: Jaro Winkler Similarity.

    Examples:
        >>> rltk.jaro_winkler_similarity('abchello', 'abcworld')
        0.6833333333333332
        >>> rltk.jaro_winkler_similarity('hello', 'world')
        0.4666666666666666
    """
    return 1 - _jaro_winkler(s1, s2, threshold, scaling_factor, prefix_len)


def jaro_distance(s1, s2):
    """
    Args:
        s1 (str): Sequence 1.
        s2 (str): Sequence 2.

    Returns:
        float: Jaro Distance.

    Examples:
        >>> rltk.jaro_distance('abc', 'abd')
        0.7777777777777777
        >>> rltk.jaro_distance('abccd', 'abcdc')
        0.9333333333333332
    """
    return _jaro_distance(s1, s2)


def _jaro_distance(s1, s2):
    # code from https://github.com/nap/jaro-winkler-distance
    # Copyright Jean-Bernard Ratte

    utils.check_for_none(s1, s2)
    utils.check_for_type(str, s1, s2)

    # s1 = utils.unicode_normalize(s1)
    # s2 = utils.unicode_normalize(s2)

    shorter, longer = s1.lower(), s2.lower()

    if len(s1) > len(s2):
        longer, shorter = shorter, longer

    m1 = _get_matching_characters(shorter, longer)
    m2 = _get_matching_characters(longer, shorter)

    if len(m1) == 0 or len(m2) == 0:
        return 0.0

    return (float(len(m1)) / len(shorter) +
            float(len(m2)) / len(longer) +
            float(len(m1) - _transpositions(m1, m2)) / len(m1)) / 3.0


def _get_diff_index(first, second):
    if first == second:
        return -1

    if not first or not second:
        return 0

    max_len = min(len(first), len(second))
    for i in range(0, max_len):
        if not first[i] == second[i]:
            return i

    return max_len


def _get_prefix(first, second):
    if not first or not second:
        return ''

    index = _get_diff_index(first, second)
    if index == -1:
        return first
    elif index == 0:
        return ''
    else:
        return first[0:index]


def _get_matching_characters(first, second):
    common = []
    limit = math.floor(min(len(first), len(second)) / 2)

    for i, l in enumerate(first):
        left, right = int(max(0, i - limit)), int(min(i + limit + 1, len(second)))
        if l in second[left:right]:
            common.append(l)
            second = second[0:second.index(l)] + '*' + second[second.index(l) + 1:]

    return ''.join(common)


def _transpositions(first, second):
    return math.floor(len([(f, s) for f, s in zip(first, second) if not f == s]) / 2.0)



class KeyValueAdapter(object):
    """
    Super class of adapter of key value stores.
    """
    def __init__(self):
        pass

    def __del__(self):
        """
        Same as :meth:`close`.
        """
        self.close()

    #: If this adapter is parallel-safe. Defaults to False if it's not overwritten in concrete class.
    parallel_safe = False

    def get(self, key: str) -> object:
        """
        Get value.
        
        Args:
            key (str): Key. 
            
        Returns:
            object:
        """
        raise NotImplementedError

    def set(self, key: str, value: object):
        """
        Set value.
        
        Args:
            key (str): Key.
            value (object): Value.
        """
        raise NotImplementedError

    def delete(self, key):
        """
        Delete value.
        
        Args:
            key (str): Key.
        """
        raise NotImplementedError

    def clean(self):
        """
        Delete all keys in adapter.
        """
        for key, _ in self:
            self.delete(key)

    def __iter__(self):
        """
        Same as :meth:`__next__`.
        """
        return self.__next__()

    def __next__(self):
        """
        Iterator of the data store. This is not required.
        
        Returns:
            iter: key, value 
        """
        pass

    def close(self):
        """
        Close handler if needed.
        """
        pass


from rltk.blocking.block import Block
from rltk.blocking.block_black_list import BlockBlackList
from rltk.blocking.block_generator import BlockGenerator
from rltk.blocking.hash_block_generator import HashBlockGenerator
from rltk.blocking.token_block_generator import TokenBlockGenerator
from rltk.blocking.canopy_block_generator import CanopyBlockGenerator
from rltk.blocking.blocking_helper import BlockingHelper



from rltk.io.adapter.key_value_adapter import KeyValueAdapter
from rltk.io.adapter.memory_key_value_adapter import MemoryKeyValueAdapter
from rltk.io.adapter.dbm_key_value_adapter import DbmKeyValueAdapter
from rltk.io.adapter.redis_key_value_adapter import RedisKeyValueAdapter
from rltk.io.adapter.hbase_key_value_adapter import HBaseKeyValueAdapter

from rltk.io.adapter.key_set_adapter import KeySetAdapter
from rltk.io.adapter.memory_key_set_adapter import MemoryKeySetAdapter
from rltk.io.adapter.redis_key_set_adapter import RedisKeySetAdapter
from rltk.io.adapter.leveldb_key_set_adapter import LevelDbKeySetAdapter



from rltk.io.adapter import KeyValueAdapter


class MemoryKeyValueAdapter(KeyValueAdapter):
    """
    Basic in-memory (dict) adapter.
    """
    def __init__(self):
        self._dict = dict()

    def get(self, key):
        return self._dict.get(key)

    def set(self, key, value: object):
        self._dict[key] = value

    def __next__(self):
        for key, value in self._dict.items():
            yield key, value

    def delete(self, key):
        del self._dict[key]

    def clean(self):
        self._dict = dict()



import itertools

from rltk.io.adapter.key_set_adapter import KeySetAdapter
from rltk.io.adapter.memory_key_set_adapter import MemoryKeySetAdapter
from rltk.dataset import Dataset
from rltk.record import Record


class Block(object):
    """
    Block
    
    key_set_adapter (keySetAdapter, optional): Where the block stores. If it's None, 
                                            :meth:`MemoryKeySetAdapter` is used. Defaults to None.
    """
    def __init__(self, key_set_adapter: KeySetAdapter = None):
        if not key_set_adapter:
            key_set_adapter = MemoryKeySetAdapter()
        self.key_set_adapter = key_set_adapter

    def add(self, block_id, dataset_id, record_id):
        """
        Add to block.
        
        Args:
            block_id (str): Block id.
            dataset_id (str / Dataset): Dataset id or Dataset object.
            record_id (str / Record): Record id or Record object.
        """
        if isinstance(dataset_id, Dataset):
            dataset_id = dataset_id.id
        if isinstance(record_id, Record):
            record_id = record_id.id
        self.key_set_adapter.add(block_id, (dataset_id, record_id))

    def get(self, block_id):
        """
        Get block by block_id.
        
        Args:
            block_id (str): Block id.
        
        Returns:
            set: {(dataset_id, record_id)}
        """
        return self.key_set_adapter.get(block_id)

    def __iter__(self):
        """
        Same as :meth:`__next__`
        """
        return self.__next__()

    def __next__(self):
        """
        Iterator of blocks.

        Returns:
            iter: block_id, dataset_id, record_id.
        """
        for block_id, data in self.key_set_adapter:
            for dataset_id, record_id in data:
                yield block_id, dataset_id, record_id

    def pairwise(self, ds_id1: str, ds_id2: str = None):
        """
        Iterator of id pairs generated according to blocks.

        Returns:
            iter: block_id, id1, id2.
        """
        if isinstance(ds_id1, Dataset):
            ds_id1 = ds_id1.id
        if ds_id2 and isinstance(ds_id2, Dataset):
            ds_id2 = ds_id2.id

        if ds_id2:
            for block_id, data in self.key_set_adapter:
                # fetch one block
                ds1, ds2 = list(), list()
                for dataset_id, record_id in data:
                    if dataset_id == ds_id1:
                        ds1.append(record_id)
                    elif dataset_id == ds_id2:
                        ds2.append(record_id)

                # cross product
                for id1, id2 in itertools.product(ds1, ds2):
                    yield block_id, id1, id2
        else:
            for block_id, data in self.key_set_adapter:
                # fetch one block
                ds1 = list()
                for dataset_id, record_id in data:
                    if dataset_id == ds_id1:
                        ds1.append(record_id)

                # combinations of two elements
                for ds1, ds1_ in itertools.combinations(ds1, 2):
                    yield ds1, ds1_



from rltk.io.adapter.key_set_adapter import KeySetAdapter


class MemoryKeySetAdapter(KeySetAdapter):
    """
    Memory key set a adapter.
    """

    def __init__(self):
        self._store = dict()

    def get(self, key):
        return self._store.get(key)

    def set(self, key, value):
        if not isinstance(value, set):
            raise ValueError('value must be a set')
        self._store[key] = value

    def add(self, key, value):
        if key not in self._store:
            self._store[key] = set()
        self._store[key].add(value)

    def remove(self, key, value):
        self._store[key].remove(value)

    def delete(self, key):
        del self._store[key]

    def clean(self):
        self._store = dict()

    def __next__(self):
        for k, v in self._store.items():
            yield k, v



from rltk.evaluation.trial import Trial
import matplotlib.pyplot as plt


class Evaluation(object):
    """
    A collection of :meth:`Trial` s. It can be used to plot chart for multiple Trials.
    
    Args:
        trial_list (list[Trial], optional): List of Trials, default is None.
    """

    def __init__(self, trial_list: list = None):
        if not trial_list:
            trial_list = []
        self.trial_list = trial_list

    def add_trial(self, trial: Trial):
        """
        Add a Trial
        
        Args:
            trial (Trial): A Trial.
        """
        self.trial_list.append(trial)

    def auc(self, x, y):
        """
        Compute AUC (Area Under Curve)
        
        Args:
            x (list): list of x coordinates
            y (list): list of y coordinates
            
        Returns:
            tuple:
                float: AUC 
                list: X-axis
                list: Y-axis
        """
        coords = sorted([(x[i], y[i]) for i in range(len(x))])
        coords_reduced = dict()
        prev = coords[0][0]
        max_y = coords[0][1]
        for c in coords:
            if prev == c[0]:
                if c[1] > max_y:
                    max_y = c[1]
            else:
                coords_reduced[prev] = max_y
                prev = c[0]
                max_y = c[1]
        coords_reduced[prev] = max_y

        x1 = 0
        y1 = 0
        auc = 0
        first = True
        for key, value in coords_reduced.items():
            if not first:
                auc += (key - x1) * (value + y1) / 2
            x1 = key
            y1 = value
            first = False
        return [auc, list(coords_reduced.keys()), list(coords_reduced.values())]

    def plot(self, parameter_list, label_max=False, label_min=False, auc_params=None, aoc_params=None,
             x_label=None, y_label=None):
        """
        General purpose plotting interface.
        
        Args:
            parameter_list (list): list of object::
            
                [{
                    'x': 'name of a property in Trial',
                    'y': 'name of a property in Trial',
                    'fmt': 'formatting of color, line style. defaults to b-',
                    'label': 'label name, this can be empty'
                }]
                
            label_max (bool, optional): whether to label max
            label_min (bool, optional): whether to label min
            auc_params (list, optional): list of AUC labelling and shading parameters::
            
                [
                    desired x coordinate of AUC label (float),
                    desired y coordinate of AUC label (float),
                    whether to shading AUC is desired (bool)
                ]
                
            aoc_params (list, optional): list of AOC labelling and shading parameters::
            
                [
                    desired x coordinate of AOC label (float),
                    desired y coordinate of AOC label (float),
                    whether to shading AOC is desired (bool)
                ]
            
            x_labels (str, optional): Label of X-axis, defaults to None.
            y_labels (str, optional): Label of Y-axis, defaults to None.
            
        Returns:
            matplotlib.pyplot:
        """

        if len(self.trial_list) == 0:
            raise Exception("Empty trial list")

        plt.figure()

        for param in parameter_list:
            x, y = [], []

            x_key, y_key = param['x'], param['y']
            fmt = param.get('fmt', 'b-')
            other_parameters = {}
            for k in param.keys():
                if k in ('x', 'y', 'fmt'):
                    continue
                other_parameters[k] = param[k]

            for trial in self.trial_list:
                x.append(getattr(trial, x_key))
                y.append(getattr(trial, y_key))

            plt.plot(x, y, fmt, **other_parameters)

        plt.legend(loc='upper right')
        if x_label:
            plt.xlabel(x_label)
        if y_label:
            plt.ylabel(y_label)

        if label_max:
            global_max = max([(x[i], y[i]) for i in range(len(x))], key=lambda i: (i[1], -i[0]))
            plt.annotate("(" + ("%.3f" % global_max[0]) + ", " + ("%.3f" % global_max[1]) + ")",
                         xy=(global_max[0] - 0.1, global_max[1] + 0.05))

        if label_min:
            global_min = min([(x[i], y[i]) for i in range(len(x))], key=lambda i: (i[1], -i[0]))
            plt.annotate("(" + ("%.3f" % global_min[0]) + ", " + ("%.3f" % global_min[1]) + ")",
                         xy=(global_min[0] - 0.1, global_min[1] - 0.05))

        if auc_params:
            vals = self.auc(x, y)
            auc = vals[0]
            area_label = 'AUC: ' + ('%.5f' % auc)
            plt.annotate(area_label, xy=(auc_params[0], auc_params[1]))

            if auc_params[2]:
                x_vals = vals[1]
                y_vals = vals[2]
                plt.fill_between(x, y)

        if aoc_params:
            vals = self.auc(x, y)
            aoc = 1 - vals[0]
            area_label = 'AOC: ' + ('%.5f' % aoc)
            plt.annotate(area_label, xy=(aoc_params[0], aoc_params[1]))

            if aoc_params[2]:
                x_vals = vals[1]
                y_vals = vals[2]
                plt.fill_between(x, y, y2=1)

        plt.xlim(0, 1.05)
        plt.ylim(0, 1.05)

        return plt

    def plot_precision_recall(self, show_points: bool = False):
        """
        Plot precision recall curve. X-axis is recall, Y-axis is precision.
        
        Args:
            show_points (bool, optional): Show points or not. Default is False.
        
        Returns:
            matplotlib.pyplot:
        """

        p_list = [{
            'x': 'recall',
            'y': 'precision',
            'fmt': 'b-',
            'label': 'Precision - Recall',
        }]
        if show_points:
            p_list.append({
                'x': 'recall',
                'y': 'precision',
                'fmt': 'bo'
            })
        return self.plot(p_list, x_label='Precision', y_label='Recall')

    def plot_roc(self):
        """
        Plot ROC (Receiver Operating Characteristic) curve. X-axis is false positives, Y-axis is true positives.
        
        Returns:
            matplotlib.pyplot:
        """
        return self.plot([{
            'x': 'false_positives',
            'y': 'true_positives',
            'label': 'ROC'
        }], auc_params=[0.05, 0.95, True], x_label='False Positives', y_label='True Negatives')

    def plot_features(self, df, f1, f2):
        """
        Plot distribution of two features.
        
        Args:
            df (pandas.DataFrame): dataframe containing trial results and computed features
            f1 (str): feature to place on the x-axis
            f2 (str): feature to place on the y-axis
            
        Returns:
            matplotlib.pyplot:
        """
        tpX = []
        tpY = []
        fpX = []
        fpY = []
        tnX = []
        tnY = []
        fnX = []
        fnY = []
        count = 0
        
        for index, row in df.iterrows():
            count += 1
            l = row['ground_truth.label']
            p = row['is_positive']
            if int(l) == 1 and p == 1:
                tpX.append(row[f1])
                tpY.append(row[f2])
            if int(l) == 0 and p == 1:
                fpX.append(row[f1])
                fpY.append(row[f2])
            if int(l) == 1 and p == 0:
                fnX.append(row[f1])
                fnY.append(row[f2])
            if int(l) == 0 and p == 0:
                tnX.append(row[f1])
                tnY.append(row[f2])
                
        plt.subplot(1, 2, 1)
        plt.plot(tpX, tpY, 'o', color='green')
        plt.plot(fpX, fpY, 'o', color='red')
        plt.xlabel(f1)
        plt.ylabel(f2)
        plt.title(f1 + ' vs ' + f2 + ' TP and FP')
        
        plt.subplot(1, 2, 2)
        plt.plot(tnX, tnY, 'o', color='green')
        plt.plot(fnX, fnY, 'o', color='red')
        plt.xlabel(f1)
        plt.ylabel(f2)
        plt.title(f1 + ' vs ' + f2 + ' TN and FN')
        plt.tight_layout()
        return plt



import time

from rltk.parallel_processor import ParallelProcessor


def test_basic():
    def dummy_computation():
        time.sleep(0.0001)

    pp = ParallelProcessor(dummy_computation, 8)
    pp.start()

    for i in range(1000):
        pp.compute()

    pp.task_done()
    pp.join()


def test_with_input():
    def dummy_computation_with_input(x):
        time.sleep(0.0001)

    pp = ParallelProcessor(dummy_computation_with_input, 8)
    pp.start()

    for i in range(1000):
        pp.compute(i)

    pp.task_done()
    pp.join()


def test_with_multiple_input():
    def dummy_computation_with_input(x, y):
        assert x * 2 == y
        time.sleep(0.0001)

    pp = ParallelProcessor(dummy_computation_with_input, 8)
    pp.start()

    for i in range(1000):
        pp.compute(i, y=i*2)

    pp.task_done()
    pp.join()


def test_with_output():
    result = []

    def dummy_computation_with_input(x):
        time.sleep(0.0001)
        return x * x

    def output_handler(r):
        result.append(r)

    pp = ParallelProcessor(dummy_computation_with_input, 8, output_handler=output_handler)
    pp.start()

    for i in range(8):
        pp.compute(i)

    pp.task_done()
    pp.join()

    for i in [0, 1, 4, 9, 16, 25, 36, 49]:
        assert i in result


def test_with_multiple_output():
    result = []

    def dummy_computation_with_input(x):
        time.sleep(0.0001)
        return x * x, x * x

    def output_handler(r1, r2):
        result.append(r1)

    pp = ParallelProcessor(dummy_computation_with_input, 8, output_handler=output_handler)
    pp.start()

    for i in range(8):
        pp.compute(i)

    pp.task_done()
    pp.join()

    for i in [0, 1, 4, 9, 16, 25, 36, 49]:
        assert i in result



import os
import pytest
import redis
import tempfile
import shutil

from rltk.record import Record
from rltk.io.adapter import *


class ConcreteRecord(Record):

    @property
    def id(self):
        return self.raw_object['id']

    @property
    def value(self):
        return self.raw_object['value']


record = ConcreteRecord(raw_object={'id': 'id1', 'value': 'value1'})


def _test_key_value_adapter(adapter):
    adapter.set(record.id, record)
    assert adapter.get(record.id).id == record.id
    assert adapter.get(record.id).value == record.value
    for rid, r in adapter:
        assert type(rid) == str
        assert rid == record.id
        assert r.id == record.id
        break

    assert adapter.get('no_such_key') is None
    adapter.clean()


def test_memory_key_value_adapter():
    adapter = MemoryKeyValueAdapter()
    _test_key_value_adapter(adapter)


def test_dbm_key_value_adapter():
    name = 'test_dbm_adapter'
    adapter = DbmKeyValueAdapter(name)
    _test_key_value_adapter(adapter)
    if os.path.exists(name + '.db'):
        os.remove(name + '.db')


def test_redis_key_value_adapter():
    try:
        adapter = RedisKeyValueAdapter('127.0.0.1', key_prefix='rltk_test_redis_key_value_adapter_')
        _test_key_value_adapter(adapter)
    except redis.exceptions.ConnectionError:
        return


def _test_key_set_adapter(adapter):
    adapter.set('a', set(['1', '2', '3']))
    assert adapter.get('a') == set(['1', '2', '3'])
    adapter.add('a', '4')
    assert adapter.get('a') == set(['1', '2', '3', '4'])
    adapter.remove('a', '4')
    assert adapter.get('a') == set(['1', '2', '3'])
    assert adapter.get('b') is None
    for k, v in adapter:
        assert type(k) == str
        assert k == 'a'
        assert v == set(['1', '2', '3'])
        break
    adapter.delete('a')
    assert adapter.get('a') is None
    adapter.set('c', set(['1', '2', '3']))
    adapter.clean()
    assert adapter.get('c') is None


def test_memory_key_set_adapter():
    adapter = MemoryKeySetAdapter()
    _test_key_set_adapter(adapter)


def test_leveldb_key_set_adapter():
    path = os.path.join(tempfile.gettempdir(), 'rltk_test_leveldb_key_set_adapter')
    adapter = LevelDbKeySetAdapter(path, name='test')
    _test_key_set_adapter(adapter)

    shutil.rmtree(path)


def test_redis_key_set_adapter():
    try:
        adapter = RedisKeySetAdapter('127.0.0.1', key_prefix='rltk_test_redis_key_set_adapter_')
        _test_key_set_adapter(adapter)
    except redis.exceptions.ConnectionError:
        return



import rltk.utils as utils


def hamming_distance(s1, s2):
    """
    Hamming distance used to measure the minimum number of substitutions required to change one sequence into the
    other.

    Args:
        s1 (str or list): Sequence 1.
        s2 (str or list): Sequence 2.

    Returns:
        int: Hamming distance between two sequences.

    Examples:
        >>> rltk.hamming_distance('ab','cd')
        2
        >>> rltk.hamming_distance([1,2,3],[3,2,3])
        1
    """

    utils.check_for_none(s1, s2)
    # utils.check_for_type(str, s1, s2)

    if len(s1) != len(s2):
        raise ValueError('Unequal length')

    return sum(c1 != c2 for c1, c2 in zip(s1, s2))


def normalized_hamming_distance(s1, s2):

    max_len = max(len(s1), len(s2))
    if max_len == 0:
        return 0

    distance = hamming_distance(s1, s2)
    return float(distance) / max_len


def hamming_similarity(s1, s2):
    """
    Hamming similarity is computed as 1 - normalized_hamming_distance.

    Args:
        s1 (str or list): Sequence 1.
        s2 (str or list): Sequence 2.

    Returns:
        float: Hamming similarity.

    Examples:
        >>> rltk.hamming_similarity('ab','cd')
        0
        >>> rltk.hamming_similarity([1,2,3],[3,2,3])
        0.666666666667
    """

    return 1 - normalized_hamming_distance(s1, s2)



import rltk.utils as utils


def _get_score(c1, c2, match, mismatch, score_table):
    """
    if there's no score found in score_table, match & mismatch will be used.
    """
    if c1 in score_table and c2 in score_table[c1]:
        return score_table[c1][c2]
    else:
        return match if c1 == c2 else mismatch


def needleman_wunsch_score(s1, s2, match=2, mismatch=-1, gap=-0.5, score_table=None):
    """
    Neeldman Wunsch score
    """
    utils.check_for_none(s1, s2)
    utils.check_for_type(str, s1, s2)

    score_table = score_table if isinstance(score_table, dict) else {}

    # s1 = utils.unicode_normalize(s1)
    # s2 = utils.unicode_normalize(s2)

    n1, n2 = len(s1), len(s2)
    if n1 == 0 and n2 == 0:
        return 0

    # construct matrix to get max score of all possible alignments
    dp = [[0] * (n2 + 1) for _ in range(n1 + 1)]
    for i in range(n1 + 1):
        for j in range(n2 + 1):
            if i == 0 and j == 0:  # [0,0]
                continue
            elif i == 0:  # most top row
                dp[i][j] = gap + dp[i][j - 1]
            elif j == 0:  # most left column
                dp[i][j] = gap + dp[i - 1][j]
            else:
                dp[i][j] = max(dp[i][j - 1] + gap,
                               dp[i - 1][j] + gap,
                               dp[i - 1][j - 1] + _get_score(s1[i - 1], s2[j - 1], match, mismatch, score_table))

    return dp[n1][n2]


def needleman_wunsch_similarity(s1, s2, match=2, mismatch=-1, gap=-0.5, score_table=None):
    """
    This Needleman Wunsch Similarity is computed as needlman_wunsch_score over maximum score of s1 and s2.

    Args:
        s1 (str): Sequence 1.
        s2 (str): Sequence 2.
        match (int, optional): Score of match.
        mismatch (int, optional): Score of mismatch.
        gap (int, optional): Gap penalty.
        score_dict (dict): Alignment score matrix. Default to None.

    Returns:
        float: Needleman Wunsch Similarity.
    """

    nm = needleman_wunsch_score(s1, s2, match, mismatch, gap, score_table)

    # score_table = {'a': {'c': 3}, 'e': {'f': 9, 'k': 1}}
    score_s1 = sum([_get_score(c1, c1, match, mismatch, score_table) for c1 in s1])
    score_s2 = sum([_get_score(c2, c2, match, mismatch, score_table) for c2 in s2])

    max_score = max(score_s1, score_s2)

    if max_score < nm:
        raise ValueError('Illegal value of score_table')

    return float(nm) / max_score



import csv

from rltk.io.reader import Reader
from rltk.io.io_utils import get_file_handler


class CSVReader(Reader):
    """
    CSV reader.
    
    Args:
        file_handler (str/io.IOBase): File name or file handler of input file.
        **kwargs: Other parameters used by `csv.DictReader <https://docs.python.org/3/library/csv.html#csv.DictReader>`_ .
    """

    def __init__(self, file_handler, **kwargs):
        self._file_handler = get_file_handler(file_handler)
        self._csv_reader = csv.DictReader(self._file_handler, **kwargs)

    def __next__(self):
        for obj in self._csv_reader:
            yield {t[0]: t[1] for t in obj.items()}

    def close(self):
        try:
            self._file_handler.close()
        except:
            pass



import rltk.utils as utils


def _jaccard_index(set1, set2):
    utils.check_for_none(set1, set2)
    utils.check_for_type(set, set1, set2)

    if len(set1) == 0 or len(set2) == 0:
        return 0

    # return float(len(set1 & set2)) / float(len(set1 | set2))

    inter_len = len(set1 & set2)
    return float(inter_len) / (len(set1) + len(set2) - inter_len)


def jaccard_index_similarity(set1, set2):
    """
    The Jaccard Index Similarity is then computed as intersection(set1, set2) / union(set1, set2).

    Args:
        set1 (set): Set 1.
        set2 (set): Set 2.

    Returns:
        float: Jaccard Index similarity.

    Examples:
        >>> rltk.jaccard_index_similarity(set(['a','b']), set(['a','c']))
        0.3333333333333333
        >>> rltk.jaccard_index_similarity(set(['a','b']), set(['c','d']))
        0.0
    """
    return _jaccard_index(set1, set2)


def jaccard_index_distance(set1, set2):
    """
    The Jaccard Index Distance is then computed as 1 - jaccard_index_similarity.

    Args:
        set1 (set): Set 1.
        set2 (set): Set 2.

    Returns:
        int: Jaccard Index Distance.
    """
    return 1 - jaccard_index_similarity(set1, set2)



from typing import Callable, TYPE_CHECKING

if TYPE_CHECKING:
    from rltk.dataset import Dataset
from rltk.blocking.block import Block
from rltk.blocking.block_black_list import BlockBlackList


class BlockGenerator(object):
    """
    Block generator.
    """

    def block(self, dataset: 'Dataset', function_: Callable = None, property_: str = None,
              block: Block = None, block_black_list: BlockBlackList = None):
        """
        Block on property or by function for dataset.
        
        Args:
            dataset (Dataset): Dataset.
            function_ (Callable): `function_(r: record)`.
            property_ (str): The property in Record object.
            block (Block): Where to write blocks. If None, a new block will be created. Defaults to None.
            block_black_list (BlockBlackList, optional): Where all blacklisted blocks are stored. Defaults to None.
                                    
        Returns:
            Block: 
        """
        block = BlockGenerator._block_args_check(function_, property_, block)
        return block

    @staticmethod
    def _block_args_check(function_, property_, block):
        if not function_ and not property_:
            raise ValueError('Invalid function or property')
        return block or Block()

    def generate(self, block1: Block, block2: Block, output_block: Block = None):
        """
        Generate block from two blocks of single dataset. 
        
        Args:
            block1 (Block): Block 1.
            block2 (Block): Block 2.
            output_block (Block): Where the output block goes. If None, a new block will be created. Defaults to None.
        
        Returns:
            Block:
        """
        block = BlockGenerator._generate_args_check(output_block)
        return block

    @staticmethod
    def _generate_args_check(block):
        return block or Block()



import re


re_record_id = re.compile(r'^[^*]{1,255}$')


class Record(object):
    """
    Record representation. Properties should be defined for further usage.
    
    Args:
        raw_object (object): Raw data which will be used to create properties.
    """

    remove_raw_object = False

    def __init__(self, raw_object):
        self.raw_object = raw_object

    @property
    def id(self):
        """
        Required property. Type has to be utf-8 string.
        """
        raise NotImplementedError

    def __eq__(self, other):
        """
        Only if both instances have the same class and id.
        
        Returns:
            bool: Equal or not.
        """
        if not isinstance(other, self.__class__):  # class should be exactly the same
            return False
        return self.id == other.id


class cached_property(property):
    """
    Decorator.
    If a Record property is decorated, the final value of it will be pre-calculated.
    """
    def __init__(self, func):
        self.func = func

    def __get__(self, obj, cls):
        """
        Args:
            obj (object): Record instance
            cls (class): Record class
        Returns:
            object: cached value
        """
        if obj is None:
            return self

        # create property if it's not there
        cached_name = self.func.__name__
        if cached_name not in obj.__dict__:
            obj.__dict__[cached_name] = self.func(obj)

        value = obj.__dict__.get(cached_name)
        return value

    def __reduce__(self):
        return cached_property.__new__, (cached_property,), {'func': self.func}


def remove_raw_object(cls):
    """
    Decorator.
    If a Record class is decorated, raw_object will be removed once all mark properties are cached.
    """
    cls.remove_raw_object = True
    return cls


def generate_record_property_cache(obj):
    """
    Generate final value on all cached_property decorated methods.
    
    Args:
        obj (Record): Record instance.
    """
    for prop_name, prop_type in obj.__class__.__dict__.items():
        if isinstance(prop_type, cached_property):
            getattr(obj, prop_name)

    validate_record(obj)

    if obj.__class__.remove_raw_object:
        del obj.__dict__['raw_object']


def validate_record(obj):
    """
    Property validator of record instance.
    
    Args:
        obj (Record): Record instance.
        
    Raises:
        TypeError: if id is not valid
    """
    if not isinstance(obj.id, str):
        raise TypeError('Id in {} should be an utf-8 encoded string.'.format(obj.__class__.__name__))
    if not re_record_id.match(obj.id):
        raise ValueError('Id is not valid')


def get_property_names(cls: type):
    """
    Get keys of property and cached_property from a record class.
    
    Args:
        cls (type): Record class
    
    Returns:
        list: Property names in class
    """
    keys = []
    for prop_name, prop_type in cls.__dict__.items():
        if not isinstance(prop_type, property) and not isinstance(prop_type, cached_property):
            continue
        keys.append(prop_name)
    return keys


@remove_raw_object
class AutoGeneratedRecord(Record):
    """
    Properties are auto generated based on the keys in `raw_object`.

    `raw_object` has to contain `id` which used as id in record.

    Args:
        raw_object (object): Raw data which will be used to create properties.
    """

    def __init__(self, raw_object):
        super().__init__(raw_object)
        for k, v in raw_object.items():
            if k == 'id':
                continue
            setattr(self, k, v)

    @cached_property
    def id(self):
        return self.raw_object['id']



import dbm

from rltk.io.adapter import KeyValueAdapter
from rltk.io.serializer import Serializer, PickleSerializer


class DbmKeyValueAdapter(KeyValueAdapter):
    """
    Python builtin `DBM <https://docs.python.org/3.6/library/dbm.html>`_ adapter.
    
    Args:
        filename (str): DBM file name.
        dbm_class (dbm): The value can be `dbm.gnu`, `dbm.ndbm` or `dbm.dumb`.
        serializer (Serializer, optional): The serializer used to serialize Record object. 
                                If it's None, `PickleSerializer` will be used. Defaults to None.
        clean (bool, optional): Clean adapters while starting. Defaults to False.
        
    Note:
        Performance drops when dataset is large.
    """
    def __init__(self, filename, dbm_class=dbm.ndbm, serializer: Serializer = None, clean: bool = False):
        if not serializer:
            serializer = PickleSerializer()
        self._db = dbm_class.open(filename, 'c')
        self._serializer = serializer

        if clean:
            self.clean()

    def get(self, key):
        v = self._db.get(key, None)
        if not v:
            return
        return self._serializer.loads(v)

    def set(self, key, value):
        self._db[key] = self._serializer.dumps(value)

    def __next__(self):
        for k in self._db.keys():
            key = k.decode('utf-8')
            yield key, self.get(key)

    def delete(self, key):
        del self._db[key]

    def close(self):
        self._db.close()



import multiprocessing

from rltk.map_reduce import MapReduce


num_of_process = int(max(multiprocessing.cpu_count() / 4, 1))


def test_map_reduce_number():

    def mapper(x):
        return x

    def reducer(r1, r2):
        return r1 + r2

    mr = MapReduce(1, mapper, reducer)
    mr.add_task(1)
    assert mr.join() == 1

    mr = MapReduce(num_of_process, mapper, reducer)
    mr.add_task(1)
    assert mr.join() == 1

    mr = MapReduce(1, mapper, reducer)
    for i in range(1, 101):
        mr.add_task(i)
    assert mr.join() == 5050

    mr = MapReduce(num_of_process, mapper, reducer)
    for i in range(1, 101):
        mr.add_task(i)
    assert mr.join() == 5050

    mr = MapReduce(num_of_process, mapper, reducer)
    for i in range(1, 100001):
        mr.add_task(i)
    assert mr.join() == 5000050000


def test_map_reduce_object():

    def mapper(k, v):
        return {k: v}

    def reducer(r1, r2):
        for k1, v1 in r1.items():
            if k1 in r2:
                r2[k1] += v1
            else:
                r2[k1] = v1
        return r2

    mr = MapReduce(1, mapper, reducer)
    for i in range(100):
        if i % 2 == 0:
            mr.add_task('a', i)
        else:
            mr.add_task('b', i)
    assert mr.join() == {'a': 2450, 'b': 2500}

    mr = MapReduce(num_of_process, mapper, reducer)
    for i in range(100):
        if i % 2 == 0:
            mr.add_task('a', i)
        else:
            mr.add_task('b', i)
    assert mr.join() == {'a': 2450, 'b': 2500}



"""
RLTK's MapReduce is a specific multiprocessing-based computing model for entity resolution.

It's different from normal MapReduce model:

- Manager fires up mapper and reducer processes simultaneously: Output of mapper is identical to reducer, \
    so reducers don't need to wait until all mappers finish.
- Data can be passed to mapper gradually: Mappers are waiting to consume data until user tells them no more new data \
    will be added.
- Reducing is not between two mapper's output (though the api to user is as this) \
    but output and context: Data pickling (serialization) and unpickling \
    (unserialization) for IPC are time consuming. As an alternation, each reducer process holds a context \
    which aggregates output in reducing step. \
    Once all output is reduced, reducing will be among contexts.
- It doesn't support shuffling and reduce-by-key.

Example::

    def mapper(x):
        time.sleep(0.0001)
        return x

    def reducer(r1, r2):
        return r1 + r2

    mr = rltk.MapReduce(8, mapper, reducer)
    for i in range(10000):
        mr.add_task(i)
    result = mr.join()
    print(result)

"""

import multiprocessing as mp
import queue
from typing import Callable
import sys
import logging


logger = logging.getLogger('RLTK - MapReduce')
logger.setLevel(logging.ERROR)
stdout_handler = logging.StreamHandler(sys.stdout)
stdout_handler.setFormatter(logging.Formatter('%(asctime)-15s %(name)s [%(levelname)s] %(message)s'))
logger.addHandler(stdout_handler)


class MapReduce(object):
    """
    Args:
        num_of_process (int): Number of mappers and reducers.
        mapper (Callable): Mapper function. The signature is `mapper(*args, **kwargs) -> object`.
        reducer (Callable): Reducer function. The signature is `reduce(object, object) -> object`.
                        `object` arguments are the returns from `mapper`s.
    """

    CMD_NO_NEW_DATA = 1  # no more new user data
    CMD_MAPPER_FINISH = 2  # mapper finished
    CMD_REDUCER_WAITING = 3  # reducer is waiting
    CMD_NO_RUNNING_MAPPER = 4  # no mapper is running
    CMD_REDUCER_AWAKE = 5  # awake a reducer
    CMD_REDUCER_KILL = 6  # kill a reducer
    CMD_REDUCER_FINISH = 7  # reducer finished

    def __init__(self, num_of_process: int, mapper: Callable, reducer: Callable):
        self._mapper_queue = mp.Queue()
        self._reducer_queue = mp.Queue()
        self._result_queue = mp.Queue()
        self._mapper_cmd_queue = [mp.Queue() for _ in range(num_of_process)]
        self._reducer_cmd_queue = [mp.Queue() for _ in range(num_of_process)]
        self._manager_cmd_queue = mp.Queue()

        self._manager_process = mp.Process(target=self._run_manager)
        self._mapper_process = [mp.Process(target=self._run_mapper, args=(i, ))
                          for i in range(num_of_process)]
        self._reducer_process = [mp.Process(target=self._run_reducer, args=(i, ))
                          for i in range(num_of_process)]

        self._mapper = mapper
        self._reducer = reducer
        self._num_of_process = num_of_process

        # start manager, mapper and reducer processes
        self._manager_process.start()
        for m in self._mapper_process:
            m.start()
        for r in self._reducer_process:
            r.start()

    def add_task(self, *args, **kwargs):
        """
        Add data.

        Args:
            args: Same to args in `mapper` function.
            kwargs: Same to kwargs in `mapper` function.
        """
        self._mapper_queue.put( (args, kwargs) )

    def join(self):
        """
        This method blocks until all mappers and reducers finish.

        Returns:
            object: The final reduced object.
        """
        # no more user data
        self._manager_cmd_queue.put( (self.__class__.CMD_NO_NEW_DATA,) )

        # reduced result
        result = self._result_queue.get()

        # make sure all child processes exited
        # (do this after clean up all queues to avoid deadlock
        # https://docs.python.org/3.6/library/multiprocessing.html?highlight=process#all-start-methods
        # "Joining processes that use queues")
        for m in self._mapper_process:
            m.join()
        for r in self._reducer_process:
            r.join()
        self._manager_process.join()

        return result

    def _run_manager(self):
        running_mapper = [1 for _ in range(self._num_of_process)]  # running mappers, 1 is running
        running_reducer = [1 for _ in range(self._num_of_process)]  # running reducers, 1 is running
        waiting_reducer = [0 for _ in range(self._num_of_process)]  # waiting reducers, 1 is waiting
        killing_reducer = [0 for _ in range(self._num_of_process)]  # killing reducers, 1 is asked to kill

        # only return the index where mask shows 1
        def apply_mask(mask):
            for idx, m in enumerate(mask):
                if m == 1:
                    yield idx

        while True:
            try:
                cmd = self._manager_cmd_queue.get(timeout=0.1)

                # no more user data, notify all mappers
                if cmd[0] == self.__class__.CMD_NO_NEW_DATA:
                    for q in self._mapper_cmd_queue:
                        q.put( (self.__class__.CMD_NO_NEW_DATA,) )

                # a mapper finished
                elif cmd[0] == self.__class__.CMD_MAPPER_FINISH:
                    idx = cmd[1]
                    running_mapper[idx] = 0
                    # notify reducers if all mappers are finished
                    if sum(running_mapper) == 0:
                        for r in self._reducer_cmd_queue:
                            r.put( (self.__class__.CMD_NO_RUNNING_MAPPER,) )

                # a reducer is waiting
                # if all reducers are waiting,
                # ask half of them to kill themselves and release held resources (context),
                # after being killed, wake up rest of the reducers
                elif cmd[0] == self.__class__.CMD_REDUCER_WAITING:
                    idx = cmd[1]
                    waiting_reducer[idx] = 1
                    logger.info('waiting reducer #%d', idx)

                    # total num of running reducers
                    running_reducer_num = len(list(apply_mask(running_reducer)))
                    logger.info('running reducer num %d', running_reducer_num)

                    # only one reducer and nothing to reduce anymore
                    if running_reducer_num == 1:
                        # kill last reducer
                        idx = next(apply_mask(running_reducer))
                        self._reducer_cmd_queue[idx].put( (self.__class__.CMD_REDUCER_KILL,) )
                        # return result to main process
                        self._result_queue.put(self._reducer_queue.get())
                        return

                    # total num of waiting reducers
                    waiting_reducer_num = len(list(filter(lambda x: x > 0,
                                    [waiting_reducer[idx] for idx in apply_mask(running_reducer)])))
                    logger.info('waiting reducer num %d', waiting_reducer_num)
                    logger.info('waiting reducer status %s', str(waiting_reducer))

                    # need to kill half of the reducers and release resources
                    if running_reducer_num == waiting_reducer_num:
                        # reset waiting reducer (for next round)
                        waiting_reducer = [0 for _ in range(self._num_of_process)]
                        # pick half of them to kill, notify these reducers
                        kill_reducer_num = running_reducer_num - int(running_reducer_num / 2)
                        notified_kill_reducer_num = 0
                        for idx in apply_mask(running_reducer):
                            self._reducer_cmd_queue[idx].put( (self.__class__.CMD_REDUCER_KILL,) )
                            killing_reducer[idx] = 1
                            notified_kill_reducer_num += 1
                            logging.info('killing reducer #%d', idx)
                            if kill_reducer_num == notified_kill_reducer_num:
                                break

                        # make sure these reducers are killed
                        while True:
                            cmd = self._manager_cmd_queue.get()
                            # other command, put it back
                            if cmd[0] != self.__class__.CMD_REDUCER_FINISH:
                                self._manager_cmd_queue.put(cmd)
                            else:
                                idx = cmd[1]
                                # reset state for killed reducer
                                running_reducer[idx] = 0
                                killing_reducer[idx] = 0
                                logger.info('reducer killed #%d', idx)

                                # all killed, wake up rest of the reducers
                                if sum(killing_reducer) == 0:
                                    for idx in apply_mask(running_reducer):
                                        logger.info('awaking reducer #%d', idx)
                                        self._reducer_cmd_queue[idx].put( (self.__class__.CMD_REDUCER_AWAKE,) )
                                    break

            except queue.Empty:
                continue

    def _run_mapper(self, idx):
        no_new_data = False

        while True:
            # cmd
            try:
                cmd = self._mapper_cmd_queue[idx].get_nowait()
                if cmd[0] == self.__class__.CMD_NO_NEW_DATA:
                    no_new_data = True
            except queue.Empty:
                pass

            # data
            try:
                data = self._mapper_queue.get(timeout=0.1)
                args, kwargs = data[0], data[1]
                result = self._mapper(*args, **kwargs)
                self._reducer_queue.put(result)
            except queue.Empty:
                # no more new data, mapper finishes
                if no_new_data:
                    self._manager_cmd_queue.put( (self.__class__.CMD_MAPPER_FINISH, idx) )
                    return
                continue

    def _run_reducer(self, idx):
        no_running_mapper = False
        context = None  # it holds result of last reducing, and can be used in next reducing

        while True:
            # cmd
            try:
                cmd = self._reducer_cmd_queue[idx].get_nowait()
                if cmd[0] == self.__class__.CMD_NO_RUNNING_MAPPER:
                    no_running_mapper = True
            except queue.Empty:
                pass

            # data
            try:
                if context is None:  # can't use "not" operator here, context could be empty
                    context = self._reducer_queue.get(timeout=0.1)

                m = self._reducer_queue.get(timeout=0.1)
                context = self._reducer(context, m)
            except queue.Empty:
                # there are still some alive mapper, wait for their output
                if not no_running_mapper:
                    continue

                # no data in reducer queue, ask manager and wait for further action
                self._manager_cmd_queue.put( (self.__class__.CMD_REDUCER_WAITING, idx) )
                cmd = self._reducer_cmd_queue[idx].get()
                # awake
                if cmd[0] == self.__class__.CMD_REDUCER_AWAKE:
                    continue
                # kill itself, put context back to reducer queue
                elif cmd[0] == self.__class__.CMD_REDUCER_KILL:
                    self._reducer_queue.put(context)
                    self._manager_cmd_queue.put( (self.__class__.CMD_REDUCER_FINISH, idx) )
                    return






import rltk.utils as utils


def string_equal(str1, str2):
    """
    Args:
        n1 (str): String 1.
        n2 (str): String 2.

    Returns:
        int: 0 for unequal and 1 for equal.
    """

    utils.check_for_none(str1, str2)
    utils.check_for_type(str, str1, str2)
    return int(str1 == str2)


def number_equal(num1, num2, epsilon=0):
    """
    Args:
        n1 (int / float): Number 1.
        n2 (int / float): Number 2.
        epsilon (float, optional): Approximation margin.

    Returns:
        int: 0 for unequal and 1 for equal.
    """

    utils.check_for_type((int, float), num1, num2)
    return int(abs(num1 - num2) <= epsilon)



from rltk.io.adapter.key_set_adapter import KeySetAdapter
from rltk.io.adapter.memory_key_set_adapter import MemoryKeySetAdapter
from rltk.blocking.block import Block


class BlockBlackList(object):
    """
    Block black list
    
    Args:
        key_set_adapter (keySetAdapter, optional): Where the block stores. If it's None, 
                                            :meth:`MemoryKeySetAdapter` is used. Defaults to None.
        max_size (int, optional): Maximum size of a block. Used by :meth:`add`. Defaults to 0.
    """
    def __init__(self, key_set_adapter: KeySetAdapter = None, max_size: int = 0):
        if not key_set_adapter:
            key_set_adapter = MemoryKeySetAdapter()
        self.key_set_adapter = key_set_adapter
        self._max_size = max_size

    def has(self, block_id: str):
        """
        Test if block_id is in black list.
        
        Args:
            block_id (str): Block id.
        """
        return self.key_set_adapter.get(block_id) is not None

    def add(self, block_id: str, block: Block):
        """
        Add block_id to black list and update block data.
        
        Args:
            block_id (str): Block id.
            block (Block): Block object.
            
        Notes:
            * If `max_size` is 0, then block_id will be added.
            * If `max_size` is greater than 0 and data in this block is more than this size,
                this block_id will be added to BlockBlackList and this block is removed from Block.
        """
        if self._max_size > 0:
            d = block.key_set_adapter.get(block_id)
            if len(d) > self._max_size:
                self.key_set_adapter.set(block_id, set())
                block.key_set_adapter.delete(block_id)
        else:
            self.key_set_adapter.set(block_id, set())

    def __contains__(self, item):
        """
        Same as :meth:`has`
        """
        self.has(item)



import rltk.utils as utils


def nysiis(s):
    """
    New York State Immunization Information System (NYSIIS) Phonetic Code is a phonetic algorithm created by `The New York State Department of Health's (NYSDOH) Bureau of Immunization
    <https://www.health.ny.gov/prevention/immunization/information_system/>`_.

    Args:
        s1 (str): Sequence 1.
        s2 (str): Sequence 2.

    Returns:
        int: 1 for same NYSIIS code, 0 for different.

    Examples:
        >>> rltk.nysiis_similarity('ashcraft', 'pineapple')
        0
    """
    # code from https://github.com/jamesturk/jellyfish
    # Copyright (c) 2015, James Turk
    # Copyright (c) 2015, Sunlight Foundation
    # All rights reserved.

    utils.check_for_none(s)
    utils.check_for_type(str, s)

    s = utils.unicode_normalize(s)

    if len(s) == 0:
        raise ValueError('Empty string')

    s = s.upper()
    key = []

    # step 1 - prefixes
    if s.startswith('MAC'):
        s = 'MCC' + s[3:]
    elif s.startswith('KN'):
        s = s[1:]
    elif s.startswith('K'):
        s = 'C' + s[1:]
    elif s.startswith(('PH', 'PF')):
        s = 'FF' + s[2:]
    elif s.startswith('SCH'):
        s = 'SSS' + s[3:]

    # step 2 - suffixes
    if s.endswith(('IE', 'EE')):
        s = s[:-2] + 'Y'
    elif s.endswith(('DT', 'RT', 'RD', 'NT', 'ND')):
        s = s[:-2] + 'D'

    # step 3 - first character of key comes from name
    key.append(s[0])

    # step 4 - translate remaining chars
    i = 1
    len_s = len(s)
    while i < len_s:
        ch = s[i]
        if ch == 'E' and i + 1 < len_s and s[i + 1] == 'V':
            ch = 'AF'
            i += 1
        elif ch in 'AEIOU':
            ch = 'A'
        elif ch == 'Q':
            ch = 'G'
        elif ch == 'Z':
            ch = 'S'
        elif ch == 'M':
            ch = 'N'
        elif ch == 'K':
            if i + 1 < len(s) and s[i + 1] == 'N':
                ch = 'N'
            else:
                ch = 'C'
        elif ch == 'S' and s[i + 1:i + 3] == 'CH':
            ch = 'SS'
            i += 2
        elif ch == 'P' and i + 1 < len(s) and s[i + 1] == 'H':
            ch = 'F'
            i += 1
        elif ch == 'H' and (s[i - 1] not in 'AEIOU' or (i + 1 < len(s) and s[i + 1] not in 'AEIOU')):
            if s[i - 1] in 'AEIOU':
                ch = 'A'
            else:
                ch = s[i - 1]
        elif ch == 'W' and s[i - 1] in 'AEIOU':
            ch = s[i - 1]

        if ch[-1] != key[-1][-1]:
            key.append(ch)

        i += 1

    key = ''.join(key)

    # step 5 - remove trailing S
    if key.endswith('S') and key != 'S':
        key = key[:-1]

    # step 6 - replace AY w/ Y
    if key.endswith('AY'):
        key = key[:-2] + 'Y'

    # step 7 - remove trailing A
    if key.endswith('A') and key != 'A':
        key = key[:-1]

    # step 8 was already done

    return key


def nysiis_similarity(s1, s2):
    """
    nysiis(s1) == nysiis(s2)

    Args:
        s1 (str): Sequence.
        s2 (str): Sequence.

    Returns:
        float: if nysiis(s1) equals to nysiis(s2)
    """
    return 1 if nysiis(s1) == nysiis(s2) else 0



import rltk.utils as utils


def dice_similarity(set1, set2):
    """
    The Dice similarity score is defined as twice the intersection of two sets divided by sum of lengths.

    Args:
        set1 (set): Set 1.
        set2 (set): Set 2.

    Returns:
        float: Dice similarity.

    Examples:
        >>> rltk.dice_similarity(set(['a', 'b']), set(['c', 'b']))
        0.5
    """

    utils.check_for_none(set1, set2)
    utils.check_for_type(set, set1, set2)

    if len(set1) == 0 or len(set2) == 0:
        return 0

    return 2.0 * float(len(set1 & set2)) / float(len(set1) + len(set2))



import io


class Reader(object):
    """
    Reader.
    """

    def __init__(self):
        pass

    def __iter__(self):
        """
        Same as :meth:`__next__`.
        """
        return self.__next__()

    def __next__(self):
        """
        Iterator.
        
        Returns:
            iter: `raw_object`. The raw_object is a dict which represents raw data of a logical row.
        """
        raise NotImplementedError

    def __del__(self):
        """
        Same as :meth:`close`
        """

    def close(self):
        """
        Close handler.
        """
        pass



import json
import random
from typing import Callable

from rltk.blocking.block_generator import BlockGenerator
from rltk.blocking.block import Block
from rltk.blocking.block_black_list import BlockBlackList


class CanopyBlockGenerator(BlockGenerator):
    """
    Canopy based block generator.
    
    Args:
        t1 (float): The loose distance.
        t2 (float): The tight distance.
        distance_metric (Callable): Compute the distance between two vectors return from :meth:`block`.
                              The signature is `distance(v1: List, v2: List) -> float`
    """
    def __init__(self, t1, t2, distance_metric):
        if t1 <= t2:
            raise ValueError('t1 should be greater than t2')
        if t2 <= 0:
            raise ValueError('t1 and t2 should greater than 0')

        self._t1 = t1
        self._t2 = t2
        self._distance_metric = distance_metric

    def block(self, dataset, function_: Callable = None, property_: str = None,
              block: Block = None, block_black_list: BlockBlackList = None):
        """
        The return of `property_` or `function_` should be a vector (list).
        """
        block = super()._block_args_check(function_, property_, block)
        for r in dataset:
            value = function_(r) if function_ else getattr(r, property_)
            k = self._encode_key(value)
            if block_black_list and block_black_list.has(k):
                continue
            if not isinstance(value, list):
                raise ValueError('Return of the function or property should be a vector (list)')
            block.add(k, dataset.id, r.id)
            if block_black_list:
                block_black_list.add(k, block)
        return block

    @staticmethod
    def _encode_key(obj):
        return json.dumps(obj)

    @staticmethod
    def _decode_key(str_):
        return json.loads(str_)

    def generate(self, block1: Block, block2: Block, output_block: Block = None):
        output_block = BlockGenerator._generate_args_check(output_block)
        dataset = []
        for key, _ in block1.key_set_adapter:
            dataset.append(self._decode_key(key))
        for key, _ in block2.key_set_adapter:
            dataset.append(self._decode_key(key))

        clusters = self._run_canopy_clustering(dataset, self._t1, self._t2, self._distance_metric)

        for c in clusters:
            for vec in c:
                key = self._encode_key(vec)
                set_ = block1.get(key)
                if set_:
                    for ds_id, rid in set_:
                        output_block.add(key, ds_id, rid)
                set_ = block2.get(key)
                if set_:
                    for ds_id, rid in set_:
                        output_block.add(key, ds_id, rid)
        return output_block

    @staticmethod
    def _run_canopy_clustering(dataset, t1, t2, distance_metric):
        """
        The algorithm proceeds as follows, using two thresholds t1 (the loose distance) and t2 (the tight distance), 
        where t1 > t2.

        1. Begin with the set of data points to be clustered.
        2. Remove a point from the set, beginning a new 'canopy' containing this point.
        3. For each point left in the set, assign it to the new canopy \
            if its distance to the first point of the canopy is less than the loose distance t1.
        4. If the distance of the point is additionally less than the tight distance t2, 
            remove it from the original set.
        5. Repeat from step 2 until there are no more data points in the set to cluster.
        """
        canopies = []
        while len(dataset) > 0:
            center_idx = random.randint(0, len(dataset) - 1)
            center_vec = dataset[center_idx]
            new_canopy = []
            delete_list = []
            del dataset[center_idx]

            for d_idx in range(len(dataset)):
                d = dataset[d_idx]
                distance = distance_metric(center_vec, d)
                if distance < t1:
                    new_canopy.append(d)
                if distance < t2:
                    delete_list.append(d_idx)

            # delete vector from dataset from backward
            for d_idx in sorted(delete_list, reverse=True):
                del dataset[d_idx]
            new_canopy.append(center_vec)  # add center
            canopies.append(new_canopy)
        return canopies



import math
import collections
import rltk.utils as utils


def cosine_similarity(vec1, vec2):
    """
    The cosine similarity between to vectors.

    Args:
        vec1 (list): Vector 1. List of integer or float.
        vec2 (list): Vector 2. List of integer or float. It should have the same length to vec1.

    Returns:
        float: Cosine similarity.

    Examples:
        >>> rltk.cosine_similarity([1, 2, 1, 3], [2, 5, 2, 3])
        0.91634193
    """

    utils.check_for_none(vec1, vec2)
    utils.check_for_type(list, vec1, vec2)
    if len(vec1) != len(vec2):
        raise ValueError('vec1 and vec2 should have same length')

    v_x_y, v_x_2, v_y_2 = 0.0, 0.0, 0.0
    for v1, v2 in zip(vec1, vec2):  # list of int / float
        v_x_y += v1 * v2
        v_x_2 += v1 * v1
        v_y_2 += v2 * v2

    return 0.0 if v_x_y == 0 else v_x_y / (math.sqrt(v_x_2) * math.sqrt(v_y_2))


def string_cosine_similarity(bag1, bag2):
    """
    The similarity between the two strings is the cosine of the angle between these two vectors representation.

    Args:
        bag1 (list): Bag1, tokenized string sequence.
        bag2 (list): Bag2, tokenized string sequence.

    Returns:
        float: Cosine similarity.
    """

    utils.check_for_none(bag1, bag2)
    utils.check_for_type(list, bag1, bag2)

    d1 = collections.Counter(bag1)
    d2 = collections.Counter(bag2)

    intersection = set(d1.keys()) & set(d2.keys())
    v_x_y = sum([d1[x] * d2[x] for x in intersection])
    v_x_2 = sum([v * v for k, v in d1.items()])
    v_y_2 = sum([v * v for k, v in d2.items()])

    return 0.0 if v_x_y == 0 else float(v_x_y) / (math.sqrt(v_x_2) * math.sqrt(v_y_2))



from rltk.io.reader.reader import Reader
from rltk.io.reader.array_reader import ArrayReader
from rltk.io.reader.dataframe_reader import DataFrameReader
from rltk.io.reader.csv_reader import CSVReader
from rltk.io.reader.jsonlines_reader import JsonLinesReader
from rltk.io.reader.ground_truth_reader import GroundTruthReader



import pickle

from rltk.io.serializer import Serializer


class PickleSerializer(Serializer):
    """
    `Pickle serializer <https://docs.python.org/3/library/pickle.html>`_ .
    """

    def loads(self, string):
        return pickle.loads(string)

    def dumps(self, obj):
        return pickle.dumps(obj)



import pytest

from rltk.record import Record
from rltk.dataset import Dataset
from rltk.io.reader.array_reader import ArrayReader
from rltk.blocking.block_black_list import BlockBlackList
from rltk.blocking.hash_block_generator import HashBlockGenerator
from rltk.blocking.token_block_generator import TokenBlockGenerator
from rltk.blocking.canopy_block_generator import CanopyBlockGenerator


class ConcreteRecord(Record):
   @property
   def id(self):
       return self.raw_object['id']

   @property
   def name(self):
       return self.raw_object['name']

   @property
   def category(self):
       return self.raw_object['category']


raw_data = [
    {'id': '1', 'name': 'apple', 'category': 'a'},
    {'id': '2', 'name': 'banana', 'category': 'a'},
    {'id': '3', 'name': 'apple & banana', 'category': 'b'},
    {'id': '4', 'name': 'pineapple', 'category': 'b'},
    {'id': '5', 'name': 'peach', 'category': 'b'},
    {'id': '6', 'name': 'coconut', 'category': 'b'}
]

ds = Dataset(reader=ArrayReader(raw_data), record_class=ConcreteRecord)


def test_hash_block_generator():
    bg = HashBlockGenerator()
    block = bg.block(ds, property_='category')
    for key, set_ in block.key_set_adapter:
        if key == 'a':
            assert set_ == set([(ds.id, '1'), (ds.id, '2')])
        elif key == 'b':
            assert set_ == set([(ds.id, '3'), (ds.id, '4'), (ds.id, '5'), (ds.id, '6')])
    block = bg.block(ds, function_=lambda r: r.category)
    for key, set_ in block.key_set_adapter:
        if key == 'a':
            assert set_ == set([(ds.id, '1'), (ds.id, '2')])
        elif key == 'b':
            assert set_ == set([(ds.id, '3'), (ds.id, '4'), (ds.id, '5'), (ds.id, '6')])

    block_black_list = BlockBlackList(max_size=2)
    block = bg.block(ds, property_='category', block_black_list=block_black_list)
    for key, set_ in block.key_set_adapter:
        assert key == 'a'
    for key, _ in block_black_list.key_set_adapter:
        assert key == 'b'


def test_token_block_generator():
    bg = TokenBlockGenerator()
    block = bg.block(ds, function_=lambda r: r.name.split(' '))
    for key, set_ in block.key_set_adapter:
        if key == 'apple':
            assert set_ == set([(ds.id, '1'), (ds.id, '3')])
        elif key == 'banana':
            assert set_ == set([(ds.id, '2'), (ds.id, '3')])

    block_black_list = BlockBlackList(max_size=1)
    block = bg.block(ds, function_=lambda r: r.name.split(' '), block_black_list=block_black_list)
    for key, set_ in block.key_set_adapter:
        assert len(set_) <= 1
    for key, _ in block_black_list.key_set_adapter:
        assert key in ('apple', 'banana')


def test_canopy_block_generator():
    bg = CanopyBlockGenerator(t1=5, t2=1, distance_metric=lambda x, y: abs(x[0] - y[0]))
    block = bg.block(ds, function_=lambda r: [ord(r.name[0].lower()) - 0x61])
    output_block = bg.generate(block, block)
    for k, _ in output_block.key_set_adapter:
        assert k in ('[1]', '[2]', '[0]', '[15]')



import json
import heapq
import random
from typing import Callable
import pandas as pd
from collections import OrderedDict
from operator import itemgetter
import copy

from rltk.utils import get_record_pairs
from rltk.io.reader import GroundTruthReader
from rltk.io.writer import GroundTruthWriter

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from rltk.dataset import Dataset


class GroundTruth(object):
    """
    Ground Truth
    
    Args:
        filename (string, optional): existing ground truth file name
        negative_if_not_exists (bool, optional): If pair is not in ground truth, it will be treated as negative pair
                                        (when you use :meth:`is_positive`, :meth:`is_negative`, :meth:`get_label`).
                                        Defaults to False.
    """
    ID1 = 'id1'
    ID2 = 'id2'
    LABEL = 'label'

    def __init__(self, filename: str = None, negative_if_not_exists: bool = False):
        self._ground_truth_data = {}
        self._gt_id1s = set([])
        self._gt_id2s = set([])
        self._negative_if_not_exists = negative_if_not_exists

        if filename:
            self.load(filename)

    def add_positive(self, id1: str, id2: str):
        """
        Add a positive ground truth. It's a syntactic sugar of :meth:`add_ground_truth`.

        Args:
            id1 (str): first id
            id2 (str): second id
        """
        self.add_ground_truth(id1, id2, True)

    def add_negative(self, id1: str, id2: str):
        """
        Add a negative ground truth. It's a syntactic sugar of :meth:`add_ground_truth`.

        Args:
            id1 (str): first id
            id2 (str): second id
        """
        self.add_ground_truth(id1, id2, False)

    def add_ground_truth(self, id1: str, id2: str, value: bool):
        """
        Add a pair to ground truth

        Args:
            id1 (str): first id
            id2 (str): second id
            value (bool): ground truth value
        """
        key = self.encode_ids(id1, id2)
        self._ground_truth_data[key] = value
        self._gt_id1s.add(id1)
        self._gt_id2s.add(id2)

    def is_member(self, id1: str, id2: str) -> bool:
        """
        Check if this pair is in ground truth

        Args:
            id1 (str): first id
            id2 (str): second id

        Returns:
            bool:
        """
        key = self.encode_ids(id1, id2)
        return key in self._ground_truth_data

    def get_label(self, id1: str, id2: str) -> bool:
        """
        Pair's label
        
        Args:
            id1 (str): first id
            id2 (str): second id
        
        Returns:
            bool: True if positive, else negative

        Raises:
            KeyError: if pair is not in ground truth
        """
        if not self.is_member(id1, id2):
            if self._negative_if_not_exists:
                return False
            raise KeyError('Not in ground truth')
        key = self.encode_ids(id1, id2)
        return self._ground_truth_data.get(key)

    def is_positive(self, id1: str, id2: str) -> bool:
        """
        If pair is positive in ground truth. It's a syntactic sugar of :meth:`get_label`.

        Args:
            id1 (str): first id
            id2 (str): second id

        Returns:
            bool:
            
        Raises:
            KeyError: if pair is not in ground truth
        """
        return self.get_label(id1, id2)

    def is_negative(self, id1: str, id2: str) -> bool:
        """
        If pair is negative in ground truth. It's a syntactic sugar of :meth:`get_label`.

        Args:
            id1 (str): first id
            id2 (str): second id

        Returns:
            bool:
            
        Raises:
            KeyError: if pair is not in ground truth
        """
        return not self.get_label(id1, id2)

    def load(self, filename: str):
        """
        Load the ground truth from file

        Args:
            filename (str): loading path
        """
        for obj in GroundTruthReader(filename):
            self._ground_truth_data[self.encode_ids(obj[self.ID1], obj[self.ID2])] = obj[self.LABEL] == 'True'

    def save(self, filename: str):
        """
        Save the ground truth to file

        Args:
            filename (str): saving path
        """
        w = GroundTruthWriter(filename)
        for k, v in self._ground_truth_data.items():
            ids = json.loads(k)
            w.write(ids[self.ID1], ids[self.ID2], v)
        w.close()

    def encode_ids(self, id1: str, id2: str):
        """
        Encode ids to key in ground truth dictionary

        Args:
            id1 (str): first id
            id2 (str): second id

        Returns:
            string:
        """
        key = json.dumps({self.ID1: id1, self.ID2: id2})
        return key

    def decode_ids(self, key: str):
        """
        Decode key in ground truth dictionary to ids

        Args:
            key (str):

        Returns:
            tuple:
                id1 (str)
                id2 (str)
        """
        obj = json.loads(key)
        return obj[self.ID1], obj[self.ID2]

    def __iter__(self):
        """
        Same as :meth:`__next__`
        """
        return self.__next__()

    def __next__(self):
        """
        Iterator
        
        Returns:
            iter: id1, id2, label
        """
        for k, v in self._ground_truth_data.items():
            id1, id2 = self.decode_ids(k)
            yield id1, id2, v

    def __len__(self):
        """
        Size of ground truth
        """
        return len(self._ground_truth_data)

    def generate_negatives(self, dataset1: 'Dataset', dataset2: 'Dataset',
                           score_function: Callable, num_of_negatives: int = -1,
                           range_in_gt: bool = False, exclude_from: 'GroundTruth' = None):
        """
        Args:
            dataset1 (Dataset): Dataset 1.
            dataset2 (Dataset): Dataset 2.
            score_function (Callable): User function, inputs are two :meth:`rltk.record.Record` s, return is a float.
            num_of_negatives (int, optional): Number of negatives to generate. 
                                            Default is -1 which will generate same number of negatives to positives.
            range_in_gt (bool, optional): The negatives will be generated within the range of ids 
                                        in ground truth if it's True,
                                        otherwise range will be the cross product of two datasets. 
                                        Default is False.
            exclude_from (GroundTruth, optional): Exclude the id pair which appears in this ground truth. 
                                            Defaults to None.
                                            This is especially useful when generating negatives for test set \
                                            meanwhile the pairs in train set need to be excluded.
        """
        num_of_negatives = len(self) if num_of_negatives == -1 else num_of_negatives
        max_heap = []

        for r1, r2 in get_record_pairs(dataset1, dataset2):
            if not self.is_member(r1.id, r2.id) and \
                    (not exclude_from or not exclude_from.is_member(r1.id, r2.id)) and \
                    (not range_in_gt or (r1.id in self._gt_id1s and r2.id in self._gt_id2s)):
                s = score_function(r1, r2)
                heapq.heappush(max_heap, (s, r1.id, r2.id))
                if len(max_heap) > num_of_negatives:
                    heapq.heappop(max_heap)

        for d in max_heap:
            r1_id, r2_id = d[1], d[2]
            self.add_negative(r1_id, r2_id)

    def generate_all_negatives(self, dataset1: 'Dataset', dataset2: 'Dataset',
                               range_in_gt: bool = False, exclude_from: 'GroundTruth' = None):
        """
        Args:
            dataset1 (Dataset): Dataset 1.
            dataset2 (Dataset): Dataset 2.
            range_in_gt (bool, optional): The negatives will be generated within the range of ids 
                                        in ground truth if it's True,
                                        otherwise range will be the cross product of two datasets. 
                                        Default is False.
            exclude_from (GroundTruth, optional): Exclude the id pair which appears in this ground truth. 
                                            Defaults to None.
                                            This is especially useful when generating negatives for test set \
                                            meanwhile the pairs in train set need to be excluded.
        """
        for r1, r2 in get_record_pairs(dataset1, dataset2):
            if not self.is_member(r1.id, r2.id) and \
                    (not exclude_from or not exclude_from.is_member(r1.id, r2.id)) and \
                    (not range_in_gt or (r1.id in self._gt_id1s and r2.id in self._gt_id2s)):
                self.add_negative(r1.id, r2.id)

    def generate_stratified_negatives(self, dataset1: 'Dataset', dataset2: 'Dataset',
                                      classify: Callable, num_of_strata: int, random_seed: int = None,
                                      num_of_negatives: int = -1,
                                      range_in_gt: bool = False, exclude_from: 'GroundTruth' = None):
        """
        Args:
            dataset1 (Dataset): Dataset 1.
            dataset2 (Dataset): Dataset 2.
            classify (Callable): User function, inputs are two :meth:`rltk.record.Record` s, 
                                return is an integer which identify which stratum the pair belongs to.
                                The return integer should be in range [0, num_of_strata).
            num_of_strata (int): Number of strata.
            random_seed (int, optional): The seed used by :py:meth:`random.seed`.
            num_of_negatives (int, optional): Number of negatives to generate. 
                                            Default is -1 which will generate same number of negatives to positives.
            range_in_gt (bool, optional): The negatives will be generated within the range of ids 
                                        in ground truth if it's True,
                                        otherwise range will be the cross product of two datasets. 
                                        Default is False.
            exclude_from (GroundTruth, optional): Exclude the id pair which appears in this ground truth. 
                                            Defaults to None.
                                            This is especially useful when generating negatives for test set \
                                            meanwhile the pairs in train set need to be excluded.
        """

        # add positives and negatives to different clusters
        strata = [{'p': [], 'n': []} for _ in range(num_of_strata)]

        # build strata
        for r1, r2 in get_record_pairs(dataset1, dataset2):
            if (range_in_gt and not (r1.id in self._gt_id1s and r2.id in self._gt_id2s)) or \
                    (exclude_from and exclude_from.is_member(r1.id, r2.id)):
                continue
            stratum_id = classify(r1, r2)
            p_n = 'p' if self.is_member(r1.id, r2.id) else 'n'
            strata[stratum_id][p_n].append((r1.id, r2.id))

        # compute weights: p / n
        strata_weights = {}
        for idx, s in enumerate(strata):
            stratum_id = str(idx)
            # nothing to pick
            if len(s['p']) == 0 or len(s['n']) == 0:
                strata_weights[stratum_id] = 0.0
                continue
            strata_weights[stratum_id] = float(len(s['p'])) / len(s['n'])

        # sorting
        sorted_strata_weights = OrderedDict(sorted(strata_weights.items(), key=itemgetter(1), reverse=True))

        # find out the number of negatives to pick from each stratum
        total_num = sum([len(s['p']) for s in strata]) if num_of_negatives == -1 else num_of_negatives
        num_to_pick_from_each_stratum = [0] * num_of_strata
        curr_strata_weights = copy.deepcopy(sorted_strata_weights)
        for stratum_id in sorted_strata_weights.keys():
            if total_num <= 0 or len(curr_strata_weights) == 0:
                break
            weight = sorted_strata_weights[stratum_id]
            idx = int(stratum_id)
            # normalize weights
            denominator = sum([w for w in curr_strata_weights.values()])
            num_to_pick_from_each_stratum[idx] = \
                min(round(total_num * weight / denominator), len(strata[idx]['n']))
            # prep for next round
            total_num -= num_to_pick_from_each_stratum[idx]
            curr_strata_weights.popitem(last=False)

        # pick negatives
        if random_seed:
            random.seed(random_seed)
        for idx, num in enumerate(num_to_pick_from_each_stratum):
            negs = random.sample(strata[idx]['n'], num)
            for n in negs:
                self.add_negative(n[0], n[1])

    def remove_negatives(self):
        """
        Remove all negatives from ground truth
        """
        keys_to_delete = []
        for key, label in self._ground_truth_data.items():
            if not label:
                keys_to_delete.append(key)
        for k in keys_to_delete:
            del self._ground_truth_data[k]

    def train_test_split(self, test_ratio: float = 0.3, random_seed: int = None):
        """
        Train test split.
        
        Args:
            test_ratio (float, optional): The ratio of test from all ground truth data. Default is 0.3.
            random_seed (int, optional): The seed used by :py:meth:`random.seed`.
        Returns:
            tuple:
                train_gt (GroudTruth)
                test_gt (GroudTruth)
        """
        if random_seed:
            random.seed(random_seed)

        train_gt, test_gt = GroundTruth(), GroundTruth()
        size = len(self)
        test_size = int(size * test_ratio)

        indices = [i for i in range(size)]
        random.shuffle(indices)
        test_indices = set(indices[:test_size])

        for idx, (id1, id2, label) in enumerate(self):
            if idx in test_indices:
                test_gt.add_ground_truth(id1, id2, label)
            else:
                train_gt.add_ground_truth(id1, id2, label)

        return train_gt, test_gt

    def generate_dataframe(self, **kwargs):
        """
        Generate data frame
        
        Returns:
            pandas.DataFrame:
        """
        columns = ['id1', 'id2', 'label']
        table = []
        for id1, id2, label in self:
            table.append([id1, id2, label])
        return pd.DataFrame(table, columns=columns, **kwargs)



# -*- coding: utf-8 -*-

import pytest

from rltk.similarity import *


@pytest.mark.parametrize('v1, v2, w, result', [
    ([1,2,3], [4,5,6], None, 5.196),
    ([1,2,3], [4,5,6], [1,3,9], 10.817)
])
def test_euclidean_distance(v1, v2, w, result):
    assert pytest.approx(euclidean_distance(v1, v2, w), 0.001) == result


@pytest.mark.parametrize('v1, v2, w, result', [
    ([1,2,3], [4,5,6], None, 9),
    ([1,2,3], [4,5,6], [1,3,9], 39)
])
def test_manhattan_distance(v1, v2, w, result):
    assert pytest.approx(manhattan_distance(v1, v2, w), 0.001) == result

@pytest.mark.parametrize('n1, n2, epsilon, equal', [
    (1, 2, 0, 0),
    (-1, -2, 0, 0),
    (3.2, 3.0, 0.15, 0),
    (3.2, 3.0, 0.3, 1)
])
def test_number_equal(n1, n2, epsilon, equal):
    if not isinstance(n1, (int, float)) or not isinstance(n2, (int, float)):
        with pytest.raises(ValueError):
            number_equal(n1, n2)
    else:
        assert number_equal(n1, n2, epsilon) == equal


@pytest.mark.parametrize('s1, s2, equal', [
    ('abc', 'abc', 1),
    ('hello', 'hello', 1),
    (None, 'ok', None),
    (1, 2, None),
    ('bar', 'foo', 0)
])
def test_string_equal(s1, s2, equal):
    if s1 is None or s2 is None:
        with pytest.raises(ValueError):
            string_equal(s1, s2)
    elif not isinstance(s1, str) or not isinstance(s2, str):
        with pytest.raises(TypeError):
            string_equal(s1, s2)
    else:
        assert string_equal(s1, s2) == equal


@pytest.mark.parametrize('s1, s2, distance', [
    ('abc', 'abc', 0),
    ('acc', 'abcd', None),
    ('testing', 'test', None),
    ('hello~', 'alloha', 5),
    ([1, 2, 3], [3, 2, 3], 1),
    ([2, 3, 1], [3, 2, 1], 2)
])
def test_hamming_distance(s1, s2, distance):
    if s1 is None or s2 is None:
        with pytest.raises(ValueError):
            hamming_distance(s1, s2)
    if len(s1) != len(s2):
        with pytest.raises(ValueError):
            hamming_distance(s1, s2)
    else:
        assert hamming_distance(s1, s2) == distance


@pytest.mark.parametrize('set1, set2, similarity', [
    (set(['data', 'science']), set(['data']), 0.667),
    (set([1, 1, 2, 3, 4]), set([2, 3, 4, 5, 6, 7, 7, 8]), 0.545)
])
def test_dice_similarity(set1, set2, similarity):
    if set1 is None or set2 is None:
        with pytest.raises(ValueError):
            dice_similarity(set1, set2)
    else:
        assert pytest.approx(dice_similarity(set1, set2), 0.001) == similarity


@pytest.mark.parametrize('s1, s2, distance, similarity', [
    ('abc', 'def', 3, 0.0),
    ('aaa', 'aaa', 0, 1.0),
    (None, 'abc', None, None),
    ('abc', None, None, None)
])
def test_levenshtein(s1, s2, distance, similarity):
    if s1 is None or s2 is None:
        with pytest.raises(ValueError):
            levenshtein_similarity(s1, s2)
        with pytest.raises(ValueError):
            levenshtein_distance(s1, s2)
    else:
        assert levenshtein_distance(s1, s2) == distance
        assert levenshtein_similarity(s1, s2) == similarity


@pytest.mark.parametrize('s1, s2, insert, delete, substitute,'
                         'insert_default, delete_default, substitute_default, distance', [
                             ('', 'abc', {'c': 50}, {}, {}, 100, 100, 100, 250),
                             ('a', 'abc', {'c': 50}, {}, {}, 100, 100, 100, 150),
                             ('ab', 'abc', {'c': 50}, {}, {}, 100, 100, 100, 50),
                             ('abc', 'abc', {'c': 50}, {}, {}, 100, 100, 100, 0),
                             ('abcd', 'abc', {}, {'d': 50}, {}, 100, 100, 100, 50),
                             ('abd', 'abc', {}, {}, {'d': {'c': 50}}, 100, 100, 100, 50),
                         ])
def test_weighted_levenshtein(s1, s2, insert, delete, substitute,
                              insert_default, delete_default, substitute_default, distance):
    if s1 is None or s2 is None:
        with pytest.raises(ValueError):
            levenshtein_distance(s1, s2, insert, delete, substitute,
                                 insert_default, delete_default, substitute_default)
    else:
        assert levenshtein_distance(s1, s2, insert, delete, substitute,
                                    insert_default, delete_default, substitute_default) == distance


@pytest.mark.parametrize('s1, s2, distance', [
    ('', '', 0),
    ('abc', '', 3),
    ('bc', 'abc', 1),
    ('fuor', 'four', 1),
    ('abcd', 'acb', 2),
    ('cape sand recycling', 'edith ann graham', 16),
    ('jellyifhs', 'jellyfish', 2),
    ('ifhs', 'fish', 2),
    ('Hello, world!', 'Hello,Â world!', 1),
])
def test_damerau_levenshtein(s1, s2, distance):
    if s1 is None or s2 is None:
        with pytest.raises(ValueError):
            damerau_levenshtein_distance(s1, s2)
    else:
        assert damerau_levenshtein_distance(s1, s2) == distance


@pytest.mark.parametrize('s1, s2, distance', [
    ('', '', 0),
    ('abc', '', 3),
    ('bc', 'abc', 1),
    ('fuor', 'four', 1),
    ('abcd', 'acb', 2),
    ('ca', 'abc', 3),
    ('ifhs', 'fish', 2),
])
def test_optimal_string_alignment_distance(s1, s2, distance):
    if s1 is None or s2 is None:
        with pytest.raises(ValueError):
            optimal_string_alignment_distance(s1, s2)
    else:
        assert optimal_string_alignment_distance(s1, s2) == distance


@pytest.mark.parametrize('s1, s2, score', [
    ('John Singer Sargent', 'John S. Sargent', 25),
    ('John Singer Sargent', 'Jane Klinger Sargent', 28.5),
    ('John Stanislaus Sargent', 'John S. Sargent', 23)
])
def test_needleman_wunsch(s1, s2, score):
    if s1 is None or s2 is None:
        with pytest.raises(ValueError):
            needleman_wunsch_score(s1, s2)
    else:
        assert needleman_wunsch_score(s1, s2) == score


@pytest.mark.parametrize('s1, s2, distance', [
    ('dixon', 'dicksonx', 0.767),
    ('martha', 'marhta', 0.944),
    ('dwayne', 'duane', 0.822),
    ('0ð00', '0ð00', 1)
])
def test_jaro_distance(s1, s2, distance):
    if s1 is None or s2 is None:
        with pytest.raises(TypeError):
            jaro_distance(s1, s2)
    else:
        assert pytest.approx(jaro_distance(s1, s2), 0.001) == distance


@pytest.mark.parametrize('s1, s2, similarity', [
    ('dixon', 'dicksonx', 0.813),
    ('martha', 'marhta', 0.961),
    ('dwayne', 'duane', 0.84),
    ('William', 'Williams', 0.975),
    ('', 'foo', 0),
    ('a', 'a', 1),
    ('abc', 'xyz', 0)
])
def test_jaro_winkler(s1, s2, similarity):
    if s1 is None or s2 is None:
        with pytest.raises(TypeError):
            jaro_winkler_similarity(s1, s2)
    else:
        assert pytest.approx(jaro_winkler_similarity(s1, s2), 0.001) == similarity


@pytest.mark.parametrize('s1, s2, distance', [
    ('', '', 0),
    ('abc', '', 3),
    ('bc', 'abc', 1),
    ('fuor', 'four', 2),
    ('abcd', 'acb', 3),
    ('jellyifhs', 'jellyfish', 4),
    ('ifhs', 'fish', 4),
])
def test_longest_common_subsequence_distance(s1, s2, distance):
    if s1 is None or s2 is None:
        with pytest.raises(ValueError):
            longest_common_subsequence_distance(s1, s2)
    else:
        assert longest_common_subsequence_distance(s1, s2) == distance


@pytest.mark.parametrize('s1, s2, distance', [
    ('', '', 1.0),
    ("ABCDEFG", "ABCDEFHJKL", 0.4),
    ('bc', 'abc', 0.33333333333333337),
    ('fuor', 'four', 0.25),
    ('abcd', 'acb', 0.5),
    ('jellyifhs', 'jellyfish', 0.2222222222222222),
    ('ifhs', 'fish', 0.5),
    ('Hello, world!', u'Hello,Â world!', 0.0714285714285714),
])
def test_metric_longest_common_subsequence(s1, s2, distance):
    if s1 is None or s2 is None:
        with pytest.raises(ValueError):
            metric_longest_common_subsequence(s1, s2)
    else:
        assert metric_longest_common_subsequence(s1, s2) == distance


@pytest.mark.parametrize('vec1, vec2, similarity', [
    ([1, 2, 1, 3], [2, 5, 2, 3], 0.916),
    ([1, 2], [2, 3], 0.992)
])
def test_cosine_similarity(vec1, vec2, similarity):
    if vec1 is None or vec2 is None:
        with pytest.raises(TypeError):
            cosine_similarity(vec1, vec2)
    else:
        assert pytest.approx(cosine_similarity(vec1, vec2), 0.001) == similarity


@pytest.mark.parametrize('bag1, bag2, similarity', [
    (['a', 'b', 'c'], ['a', 'a', 'c'], 0.775),
    (['cc', 'dd', 'a'], ['dd', 'a', 'cc'], 1.0)
])
def test_string_cosine_similarity(bag1, bag2, similarity):
    if bag1 is None or bag2 is None:
        with pytest.raises(TypeError):
            string_cosine_similarity(bag1, bag2)
    else:
        assert pytest.approx(string_cosine_similarity(bag1, bag2), 0.001) == similarity


@pytest.mark.parametrize('set1, set2, similarity', [
    (set(['abc', 'bcd', 'cde']), set(['cde', 'efg', 'fgh']), 0.2),
    (set(['abc', 'def']), set(['abc', 'def']), 1.0),
    (set(['abc', 'def']), set(['hij', 'klm']), 0.0)
])
def test_jaccard_index_similarity(set1, set2, similarity):
    if set1 is None or set2 is None:
        with pytest.raises(TypeError):
            jaccard_index_similarity(set1, set2)
    else:
        assert pytest.approx(jaccard_index_similarity(set1, set2), 0.001) == similarity


@pytest.mark.parametrize('bag1, bag2, df_corpus, doc_size, math_log, score', [
    # (['a', 'b', 'a'], ['a', 'c'], [['a', 'b', 'a'], ['a', 'c'], ['a']], False, 0.1754),
    # (['a', 'b', 'a'], ['a', 'c'], [['a', 'b', 'a'], ['a', 'c'], ['a'], ['b']], True, 0.1117),
    # (['a', 'b', 'a'], ['a'], [['a', 'b', 'a'], ['a', 'c'], ['a']], False, 0.5547),
    # (['a', 'b', 'a'], ['a'], [['x', 'y'], ['w'], ['q']], False, 0.0),
    # (['a', 'b', 'a'], ['a'], [['x', 'y'], ['w'], ['q']], True, 0.0),
    # (['a', 'b', 'a'], ['a'], None, False, 0.7071)
    (['a', 'b', 'a'], ['a', 'c'], {'a': 3, 'b': 1, 'c': 1}, 3, False, 0.1754),
    (['a', 'b', 'a'], ['a', 'c'], {'a': 3, 'b': 2, 'c': 1}, 4, True, 0.1297),
    (['a', 'b', 'a'], ['a'], {'a': 3, 'b': 1, 'c': 1}, 3, False, 0.5547),
    (['a', 'b', 'a'], ['a'], {'x': 1, 'y': 1, 'w': 1, 'q': 1}, 3, False, 0.0),
    (['a', 'b', 'a'], ['a'], {'x': 1, 'y': 1, 'w': 1, 'q': 1}, 3, True, 0.0),
    (['a', 'b', 'a'], ['a'], None, 0, False, 0.7071)
])
def test_tf_idf(bag1, bag2, df_corpus, doc_size, math_log, score):
    if bag1 is None or bag2 is None or df_corpus is None:
        with pytest.raises(TypeError):
            tf_idf_similarity(bag1, bag2)
    else:
        assert pytest.approx(tf_idf_similarity(bag1, bag2, df_corpus, doc_size, math_log), 0.001) == score


@pytest.mark.parametrize('s1, s2, n, distance', [
    ('', '', 2, 1),
    ('abcd', 'abcd', 2, 0),
    ("ABCDEFG", "ABCDEFHJKL", 2, 0.35),
    ('bc', 'abc', 2, 0.5),
    ('fuor', 'four', 2, 0.5),
    ('abcd', 'acb', 2, 0.5),
    ('jellyifhs', 'jellyfish', 2, 0.3888888888888889),
    ('ifhs', 'fish', 2, 0.875),
    ('abcdef', 'abdcef', 3, 0.3333333333333333),
])
def test_ngram_distance(s1, s2, n, distance):
    if s1 is None or s2 is None:
        with pytest.raises(ValueError):
            ngram_distance(s1, s2 ,n)
    else:
        assert ngram_distance(s1, s2 ,n) == distance


@pytest.mark.parametrize('s1, s2, n, similarity', [
    ('', '', 2, 0),
    ('abcd', 'abcd', 2, 1),
    ("ABCDEFG", "ABCDEFHJKL", 2, 0.65),
    ('bc', 'abc', 2, 0.6666666666666666),
    ('fuor', 'four', 2, 0.5),
    ('abcd', 'acb', 2, 0.5),
    ('jellyifhs', 'jellyfish', 2, 0.7222222222222222),
    ('ifhs', 'fish', 2, 0.5),
    ('abcdef', 'abdcef', 3, 0.6666666666666666),
])
def test_ngram_similarity(s1, s2, n, similarity):
    if s1 is None or s2 is None:
        with pytest.raises(ValueError):
            ngram_distance(s1, s2, n)
    else:
        assert ngram_similarity(s1, s2, n) == similarity


@pytest.mark.parametrize('s1, s2, n, distance', [
    ('', '', 2, 1),
    ('abcd', 'abcd', 2, 0),
    ("ABCDEFG", "ABCDEFHJKL", 2, 5),
    ('bc', 'abc', 2, 1),
    ('fuor', 'four', 2, 6),
    ('abcd', 'acb', 2, 5),
    ('jellyifhs', 'jellyfish', 2, 8),
    ('ifhs', 'fish', 2, 6),
    ('abcdef', 'abdcef', 3, 8),
])
def test_qgram_distance(s1, s2, n, distance):
    if s1 is None or s2 is None:
        with pytest.raises(ValueError):
            qgram_distance(s1, s2, n)
    else:
        assert qgram_distance(s1, s2, n) == distance


@pytest.mark.parametrize('s1, s2, n, similarity', [
    ('', '', 2, 0),
    ('abcd', 'abcd', 2, 3),
    ("ABCDEFG", "ABCDEFHJKL", 2, 5),
    ('bc', 'abc', 2, 1),
    ('fuor', 'four', 2, 0),
    ('abcd', 'acb', 2, 0),
    ('jellyifhs', 'jellyfish', 2, 4),
    ('ifhs', 'fish', 2, 0),
    ('abcdef', 'abdcef', 3, 0),
])
def test_qgram_similarity(s1, s2, n, similarity):
    if s1 is None or s2 is None:
        with pytest.raises(ValueError):
            qgram_similarity(s1, s2, n)
    else:
        assert qgram_similarity(s1, s2, n) == similarity


def test_hybrid_jaccard_similarity():
    # use a fixed test cases here only to test hybrid jaccard itself.
    def test_function(m, n):
        if m == 'a' and n == 'p':
            return 0.7
        if m == 'a' and n == 'q':
            return 0.8
        if m == 'b' and n == 'p':
            return 0.5
        if m == 'b' and n == 'q':
            return 0.9
        if m == 'c' and n == 'p':
            return 0.2
        if m == 'c' and n == 'q':
            return 0.1

    assert pytest.approx(hybrid_jaccard_similarity(set(['a', 'b', 'c']), set(['p', 'q']), function=test_function),
                         0.001) == 0.5333


@pytest.mark.parametrize('bag1, bag2, similarity', [
    (['paul', 'johnson'], ['johson', 'paule'], 0.944),
    (['Niall'], ['Neal'], 0.805)
])
def test_monge_elkan_similarity(bag1, bag2, similarity):
    if bag1 is None or bag2 is None:
        with pytest.raises(TypeError):
            monge_elkan_similarity(bag1, bag2)
    else:
        assert pytest.approx(monge_elkan_similarity(bag1, bag2), 0.001) == similarity


@pytest.mark.parametrize('s, code', [
    ('Soundex', 'S532'),
    ('Example', 'E251'),
    ('Sownteks', 'S532'),
    ('Ekzampul', 'E251'),
    ('Euler', 'E460'),
    ('Gauss', 'G200'),
    ('Hilbert', 'H416'),
    ('Knuth', 'K530'),
    ('Lloyd', 'L300'),
    ('Lukasiewicz', 'L222'),
    ('Ellery', 'E460'),
    ('Ghosh', 'G200'),
    ('Heilbronn', 'H416'),
    ('Kant', 'K530'),
    ('Ladd', 'L300'),
    ('Lissajous', 'L222'),
    ('Wheaton', 'W350'),
    ('Burroughs', 'B620'),
    ('Burrows', 'B620'),
    ('O\'Hara', 'O600'),
    ('Washington', 'W252'),
    ('Lee', 'L000'),
    ('Gutierrez', 'G362'),
    ('Pfister', 'P236'),
    ('Jackson', 'J250'),
    ('Tymczak', 'T522'),
    ('VanDeusen', 'V532'),
    ('Ashcraft', 'A261'),
    ('Gutierrez', 'G362'),
    ('Çáŕẗéř', 'C636'),
])
def test_soundex(s, code):
    if s is None:
        with pytest.raises(ValueError):
            soundex(s)
    if isinstance(s, int):
        with pytest.raises(TypeError):
            soundex(s)
    else:
        assert soundex(s) == code


@pytest.mark.parametrize('s, code', [
    ('DGIB', 'JB'),
    ('metaphone', 'MTFN'),
    ('wHErE', 'WR'),
    ('shell', 'XL'),
    ('this is a difficult string', '0S IS A TFKLT STRNK'),
    ('aeromancy', 'ERMNS'),
    ('Antidisestablishmentarianism', 'ANTTSSTBLXMNTRNSM'),
    ('sunlight labs', 'SNLT LBS'),
    ('sonlite laabz', 'SNLT LBS'),
    ('Çáŕẗéř', 'KRTR'),
    ('kentucky', 'KNTK'),
    ('KENTUCKY', 'KNTK'),
    ('NXNXNX', 'NKSNKSNKS'),
    ('Aapti', 'PT'),
    ('Aarti', 'RT'),
    ('CIAB', 'XB'),
    ('NQ', 'NK'),
    ('sian', 'XN'),
    ('gek', 'JK'),
    ('Hb', 'HB'),
    ('Bho', 'BH'),
    ('Tiavyi', 'XFY'),
    ('Xhot', 'XHT'),
    ('Xnot', 'SNT'),
    ('g', 'K'),
    ('8 queens', 'KNS'),
    ('Utah', 'UT'),
    ('WH', 'W')
])
def test_metaphone(s, code):
    if s is None:
        with pytest.raises(ValueError):
            metaphone(s)
    if isinstance(s, int):
        with pytest.raises(TypeError):
            metaphone(s)
    else:
        assert metaphone(s) == code


@pytest.mark.parametrize('s, code', [
    ('Worthy', 'WARTY'),
    ('Ogata', 'OGAT'),
    ('montgomery', 'MANTGANARY'),
    ('Costales', 'CASTAL'),
    ('Tu', 'T'),
    ('martincevic', 'MARTANCAFAC'),
    ('Catherine', 'CATARAN'),
    ('Katherine', 'CATARAN'),
    ('Katerina', 'CATARAN'),
    ('Johnathan', 'JANATAN'),
    ('Jonathan', 'JANATAN'),
    ('John', 'JAN'),
    ('Teresa', 'TARAS'),
    ('Theresa', 'TARAS'),
    ('Jessica', 'JASAC'),
    ('Joshua', 'JAS'),
    ('Bosch', 'BAS'),
    ('Lapher', 'LAFAR'),
    ('wiyh', 'WY'),
    ('MacArthur', 'MCARTAR'),
    ('Pheenard', 'FANAD'),
    ('Schmittie', 'SNATY'),
    ('Knaqze', 'NAGS'),
    ('Knokno', 'NAN'),
    ('Knoko', 'NAC'),
    ('Macaw', 'MC'),
    ('T', 'T'),
    ('S', 'S'),
    ('P', 'P'),
    ('K', 'C'),
    ('M', 'M'),
    ('E', 'E')
])
def test_nysiis(s, code):
    if s is None:
        with pytest.raises(ValueError):
            nysiis(s)
    if isinstance(s, int):
        with pytest.raises(TypeError):
            nysiis(s)
    else:
        assert nysiis(s) == code



import os
import plyvel

from rltk.io.serializer import Serializer, PickleSerializer
from rltk.io.adapter.key_set_adapter import KeySetAdapter


class LevelDbKeySetAdapter(KeySetAdapter):
    """
    `LevelDB <https://github.com/google/leveldb>`_ key set adapter. 
    LevelDB is a serverless, stand-alone key value store. It can be used as a local file system store.
    
    
    Args:
        path (str): The directory path used by LevelDB.
        name (str): Because LevelDB only has a single key space, \
                    this is used as name space.
        serializer (Serializer, optional): The serializer used to serialize each object in set. 
                                If it's None, `PickleSerializer` will be used. Defaults to None.
        clean (bool, optional): Clean adapters while starting. Defaults to False.
        kwargs: Other key word arguments for `plyvel.DB <https://plyvel.readthedocs.io/en/latest/api.html#DB>`_.
        
    Note:
        A particular LevelDB database only supports accessing by one process at one time. 
        This adapter uses singleton (in one RLTK instance) to make sure only one `plyvel.DB` is created.
        Different `name` s can be used if you don't want to create multiple databases.
    """
    _db_instance = None
    _db_ref_count = 0

    def __init__(self, path: str, name: str, serializer: Serializer = None, clean: bool = False, **kwargs):
        if not serializer:
            serializer = PickleSerializer()

        # leveldb's connection can only be a singleton
        if not self.__class__._db_instance:
            if not os.path.exists(path):
                os.mkdir(path)
            self.__class__._db_instance = plyvel.DB(path, create_if_missing=True, **kwargs)
        self._db = self.__class__._db_instance
        self.__class__._db_ref_count += 1

        self._prefix = '{name}_'.format(name=name)
        self._prefix_db = self._db.prefixed_db(self._encode(self._prefix))
        self._serializer = serializer

        if clean:
            self.clean()

    @staticmethod
    def _encode(string):
        return string.encode(encoding='utf-8')

    @staticmethod
    def _decode(bytes_):
        return bytes_.decode(encoding='utf-8')

    def _get(self, key):
        v = self._prefix_db.get(key)
        if not v:
            return
        return self._serializer.loads(v)

    def get(self, key):
        return self._get(self._encode(key))

    def set(self, key, value):
        if not isinstance(value, set):
            raise ValueError('value must be a set')
        self.delete(key)
        self._prefix_db.put(self._encode(key), self._serializer.dumps(value))

    def add(self, key, value):
        set_ = self.get(key)
        if not set_:
            set_ = set([])
        set_.add(value)
        return self.set(key, set_)

    def remove(self, key, value):
        set_ = self.get(key)
        if not set_:
            return
        set_.remove(value)
        return self.set(key, set_)

    def delete(self, key):
        return self._prefix_db.delete(self._encode(key))

    def __next__(self):
        for key in self._prefix_db.iterator(include_value=False):
            yield self._decode(key), self._get(key)

    def close(self):
        self.__class__._db_ref_count -= 1
        if self.__class__._db_ref_count == 0:
            self._db.close()



from collections import defaultdict

import rltk.utils as utils


def levenshtein_distance(s1, s2, insert=None, delete=None, substitute=None,
                         insert_default=1, delete_default=1, substitute_default=1):
    """
    The Levenshtein distance between two words is the minimum number of single-character edits (insertions, deletions or substitutions) required to change one word into the other.

    Args:
        s1 (str): Sequence 1.
        s2 (str): Sequence 2.
        insert (dict(str, int), optional): Insert cost of characters. Defaults to None.
        delete (dict(str, int), optional): Delete cost of characters. Defaults to None.
        substitute (dict(str, dict(str, int)), optional): Substitute cost of characters. Defaults to None.
        insert_default (int, optional): Default value of insert cost. Defaults to 1.
        delete_default (int, optional): Default value of delete cost. Defaults to 1.
        substitute_default (int, optional): Default value of substitute cost. Defaults to 1.

    Returns:
        int: Levenshtein Distance.

    Examples:
        >>> rltk.levenshtein_distance('ab', 'abc')
        1
        >>> rltk.levenshtein_distance('a', 'abc', insert = {'c':50},
        ... insert_default=100, delete_default=100, substitute_default=100)
        150
    """

    utils.check_for_none(s1, s2)
    utils.check_for_type(str, s1, s2)

    insert = insert if isinstance(insert, dict) else {}
    delete = delete if isinstance(delete, dict) else {}
    substitute = substitute if isinstance(substitute, dict) else {}

    # s1 = utils.unicode_normalize(s1)
    # s2 = utils.unicode_normalize(s2)

    n1, n2 = len(s1), len(s2)
    if n1 == 0 and n2 == 0:
        return 0

    # if n1 == 0 or n2 == 0:
    #     return max(n1, n2)

    dp = [[0] * (n2 + 1) for _ in range(n1 + 1)]
    for i in range(n1 + 1):
        for j in range(n2 + 1):
            if i == 0 and j == 0:  # [0,0]
                continue
            elif i == 0:  # most top row
                c = s2[j - 1]
                dp[i][j] = insert[c] if c in insert else insert_default
                dp[i][j] += dp[i][j - 1]
            elif j == 0:  # most left column
                c = s1[i - 1]
                dp[i][j] = delete[c] if c in delete else delete_default
                dp[i][j] += dp[i - 1][j]
            else:
                c1, c2 = s1[i - 1], s2[j - 1]
                insert_cost = insert[c2] if c2 in insert else insert_default
                delete_cost = delete[c1] if c1 in delete else delete_default
                substitute_cost = substitute[c1][c2] \
                    if c1 in substitute and c2 in substitute[c1] else substitute_default

                if c1 == c2:
                    dp[i][j] = dp[i - 1][j - 1]
                else:
                    dp[i][j] = min(dp[i][j - 1] + insert_cost,
                                   dp[i - 1][j] + delete_cost,
                                   dp[i - 1][j - 1] + substitute_cost)
    return dp[n1][n2]


def levenshtein_similarity(s1, s2, insert=None, delete=None, substitute=None,
                           insert_default=1, delete_default=1, substitute_default=1):
    """
    Computed as 1 - normalized_levenshtein_distance.
    """
    return 1.0 - normalized_levenshtein_distance(s1, s2, insert, delete, substitute,
                                                 insert_default, delete_default, substitute_default)


def normalized_levenshtein_distance(s1, s2, insert=None, delete=None, substitute=None,
                                    insert_default=1, delete_default=1, substitute_default=1):
    """
    Computed as levenshtein - max-insert-cost(s1,s2)
    """

    insert = insert if isinstance(insert, dict) else {}
    delete = delete if isinstance(delete, dict) else {}
    substitute = substitute if isinstance(substitute, dict) else {}

    def compute_insert_cost(s):
        cost = 0
        for c in s:
            cost += insert[c] if c in insert else insert_default
        return cost

    lev = levenshtein_distance(s1, s2, insert, delete, substitute,
                               insert_default, delete_default, substitute_default)

    max_cost = max(compute_insert_cost(s1), compute_insert_cost(s2))

    if max_cost < lev:
        raise ValueError('Illegal value of operation cost')

    if max_cost == 0:
        return 0

    return float(lev) / max_cost


def damerau_levenshtein_distance(s1, s2):
    """
    Similar to Levenshtein, Damerau-Levenshtein distance is the minimum number of operations needed to transform one string into the other, where an operation is defined as an insertion, deletion, or substitution of a single character, or a transposition of two adjacent characters.

    Args:
        s1 (str): Sequence 1.
        s2 (str): Sequence 2.

    Returns:
        float: Damerau Levenshtein Distance.

    Examples:
        >>> rltk.damerau_levenshtein_distance('abcd', 'acbd')
        1
        >>> rltk.damerau_levenshtein_distance('abbd', 'acad')
        2
    """

    utils.check_for_none(s1, s2)
    utils.check_for_type(str, s1, s2)

    # s1 = utils.unicode_normalize(s1)
    # s2 = utils.unicode_normalize(s2)

    n1, n2 = len(s1), len(s2)
    infinite = n1 + n2

    char_arr = defaultdict(int)
    dp = [[0] * (n2 + 2) for _ in range(n1 + 2)]

    dp[0][0] = infinite
    for i in range(0, n1 + 1):
        dp[i + 1][0] = infinite
        dp[i + 1][1] = i
    for i in range(0, n2 + 1):
        dp[0][i + 1] = infinite
        dp[1][i + 1] = i

    for i in range(1, n1 + 1):
        db = 0
        for j in range(1, n2 + 1):
            i1 = char_arr[s2[j - 1]]
            j1 = db
            cost = 1
            if s1[i - 1] == s2[j - 1]:
                cost = 0
                db = j

            dp[i + 1][j + 1] = min(dp[i][j] + cost,
                                   dp[i + 1][j] + 1,
                                   dp[i][j + 1] + 1,
                                   dp[i1][j1] + (i - i1 - 1) + 1 + (j - j1 - 1))
        char_arr[s1[i - 1]] = i

    return dp[n1 + 1][n2 + 1]


def optimal_string_alignment_distance(s1, s2):
        """
        This is a variation of the Damerau-Levenshtein distance that returns the strings' edit distance
        taking into account deletion, insertion, substitution, and transposition, under the condition
        that no substring is edited more than once.

        Args:
            s1 (str): Sequence 1.
            s2 (str): Sequence 2.

        Returns:
            float: Optimal String Alignment Distance.

        Examples:
            >>> rltk.optimal_string_alignment_distance('abcd', 'acbd')
            1
            >>> rltk.optimal_string_alignment_distance('ca', 'abc')
            3
        """

        utils.check_for_none(s1, s2)
        utils.check_for_type(str, s1, s2)

        # s1 = utils.unicode_normalize(s1)
        # s2 = utils.unicode_normalize(s2)

        n1, n2 = len(s1), len(s2)

        dp = [[0] * (n2 + 1) for _ in range(n1 + 1)]

        for i in range(0, n1 + 1):
            dp[i][0] = i
        for j in range(0, n2 + 1):
            dp[0][j] = j

        for i in range(1, n1 + 1):
            for j in range(1, n2 + 1):
                cost = 0 if s1[i - 1] == s2[j - 1] else 1

                dp[i][j] = min(dp[i][j - 1] + 1,
                               dp[i - 1][j] + 1,
                               dp[i - 1][j - 1] + cost)

                if (i > 1 and j > 1 and s1[i - 1] == s2[j - 2] and s1[i - 2] == s2[j - 1]):
                    dp[i][j] = min(dp[i][j], dp[i - 2][j - 2] + cost)

        return dp[n1][n2]


from scipy.optimize import linear_sum_assignment
import rltk.utils as utils
from rltk.similarity.jaro import jaro_winkler_similarity

MIN_FLOAT = float('-inf')


def hybrid_jaccard_similarity(set1, set2, threshold=0.5, function=jaro_winkler_similarity, parameters=None):
    """
    Generalized Jaccard Measure.

    Args:
        set1 (set): Set 1.
        set2 (set): Set 2.
        threshold (float, optional): The threshold to keep the score of similarity function. \
            Defaults to 0.5.
        function (function, optional): The reference of a similarity measure function. \
            It should return the value in range [0,1]. If it is set to None, \
            `jaro_winlker_similarity` will be used.
        parameters (dict, optional): Other parameters of function. Defaults to None.

    Returns:
        float: Hybrid Jaccard similarity.

    Examples:
        >>> def hybrid_test_similarity(m ,n):
        ...     ...
        >>> rltk.hybrid_jaccard_similarity(set(['a','b','c']), set(['p', 'q']), function=hybrid_test_similarity)
        0.533333333333
    """

    utils.check_for_none(set1, set2)
    utils.check_for_type(set, set1, set2)

    parameters = parameters if isinstance(parameters, dict) else {}

    matching_score = []
    for s1 in set1:
        inner = []
        for s2 in set2:
            score = function(s1, s2, **parameters)
            if score < threshold:
                score = 0.0
            inner.append(1.0 - score)  # munkres finds out the smallest element
        matching_score.append(inner)

    row_idx, col_idx = linear_sum_assignment(matching_score)

    score_sum, matching_count = 0.0, 0
    for r, c in zip(row_idx, col_idx):
        matching_count += 1
        score_sum += 1.0 - matching_score[r][c]  # go back to similarity

    if len(set1) + len(set2) - matching_count == 0:
        return 1.0
    return float(score_sum) / float(len(set1) + len(set2) - matching_count)


def monge_elkan_similarity(bag1, bag2, function=jaro_winkler_similarity, parameters=None):
    """
    Monge Elkan similarity.

    Args:
        bag1 (list): Bag 1.
        bag2 (list): Bag 2.
        function (function, optional): The reference of a similarity measure function. \
            It should return the value in range [0,1]. If it is set to None, \
            `jaro_winlker_similarity` will be used.
        parameters (dict, optional): Other parameters of function. Defaults to None.

    Returns:
        float: Monge Elkan similarity.
    """

    utils.check_for_none(bag1, bag2)
    utils.check_for_type(list, bag1, bag2)

    parameters = parameters if isinstance(parameters, dict) else {}

    if len(bag1) == 0:
        return 0.0

    score_sum = 0
    for ele1 in bag1:
        max_score = MIN_FLOAT
        for ele2 in bag2:
            max_score = max(max_score, function(ele1, ele2, **parameters))
        score_sum += max_score

    return float(score_sum) / float(len(bag1))


def symmetric_monge_elkan_similarity(bag1, bag2, function=jaro_winkler_similarity, parameters=None):
    """
    Symmetric Monge Elkan similarity is computed by \
    (monge_elkan_similarity(b1, b2) + monge_elkan_similarity(b2, b1)) / 2.
    """

    s1 = monge_elkan_similarity(bag1, bag2, function, parameters)
    s2 = monge_elkan_similarity(bag2, bag1, function, parameters)
    return (s1 + s2) / 2



import heapq
import pandas as pd
import copy

from rltk.record import Record, get_property_names
from rltk.evaluation.ground_truth import GroundTruth
from scipy.optimize import linear_sum_assignment
from typing import Any


class Trial(object):
    """
    Trial stores the calculated result for further evaluation.
    It only saves the result which is also in ground truth.
    
    Args:
        ground_truth (GroundTruth): Ground truth.
        min_confidence (float): If the result has lower confidence than min confidence, it will not be saved.
                                Default is 0.
        top_k (int): Max number of result to be saved. 0 means no limitation.
                    Default is 0.
        **kwargs: Other user-defined properties.
    """

    class Result:
        """
        Result data structure

        Args:
            record1 (Record): first record
            record2 (Record): second record
            is_positive (bool): if the prediction of these two records is pair
            confidence (float): the probability of positive
        """

        def __init__(self, record1: Record, record2: Record, is_positive: bool, confidence: float = None, **kwargs):
            self.record1 = record1
            self.record2 = record2
            self.is_positive = is_positive
            self.confidence = confidence
            self.extra_key_values = kwargs

        def __deepcopy__(self, memo):
            # Notice: record objects are NOT deep copied. Only reference is used here.
            cls = self.__class__
            copied = cls(record1=self.record1, record2=self.record2, is_positive=self.is_positive,
                         confidence=self.confidence)
            copied.extra_key_values = copy.deepcopy(self.extra_key_values)
            return copied

        def __cmp__(self, other):
            return self.confidence < other.confidence

        def __lt__(self, other):
            return self.confidence < other.confidence

        def __getattr__(self, key):
            return self.extra_key_values[key]

        def get_property_names(self):
            """
            Get all properties in Result
            
            Returns:
                list:
            """
            return ['is_positive', 'confidence'] + list(self.extra_key_values.keys())
            
    def __init__(self, ground_truth: GroundTruth, min_confidence: float = 0, top_k: int = 0, **kwargs):
        self._ground_truth = ground_truth
        self._min_confidence = min_confidence
        self._top_k = top_k
        self._results = []
        self.extra_key_values = kwargs
        self.pre_evaluate()

    def __deepcopy__(self, memo):
        """
        Deep copy of trial and trial results.
        
        Args:
            memo (dict): Argument `memo <https://docs.python.org/3.6/library/copy.html>`_ of :py:meth:`copy.deepcopy`.
            
        Note:
            `Record` in `Result` is still reference.
            
        Returns:
            Trial: Cloned object.
        """
        cls = self.__class__
        copied = cls(ground_truth=self._ground_truth, min_confidence=self._min_confidence, top_k=self._top_k)
        copied._results = copy.deepcopy(self._results)
        copied._extra_key_value = copy.deepcopy(self.extra_key_values)
        copied.tp = self.tp
        copied.tn = self.tn
        copied.fp = self.fp
        copied.fn = self.fn
        copied.tp_list = self.tp_list
        copied.tn_list = self.tn_list
        copied.fp_list = self.fp_list
        copied.fn_list = self.fn_list
        return copied

    def clone(self):
        """
        Same as :meth:`__deepcopy__`.
        """
        return copy.deepcopy(self)

    def add_property(self, key: str, value: Any):
        """
        Add new property to Trial
        
        Args:
            key (str): Key name
            value (Any): Any type of value
        """
        self.extra_key_values[key] = value

    def __getattr__(self, key):
        try:
            return self.extra_key_values[key]
        except:
            raise AttributeError

    def __iter__(self):
        """
        Same as :meth:`__next__`
        """
        return self.__next__()

    def __next__(self):
        """
        Iterator
        
        Returns:
            iter: Result
        """
        for r in self._results:
            yield r

    def pre_evaluate(self):
        """
        Preparation before evaluation
        """
        self.tp = 0
        self.tn = 0
        self.fp = 0
        self.fn = 0
        self.tp_list = []
        self.tn_list = []
        self.fp_list = []
        self.fn_list = []

    def evaluate(self, threshold: float = None):
        """
        Run evaluation
        
        Args:
            threshold (float, optional): Only if :meth:`Result.confidence` is greater than this threshold,
                                    `Result.is_positive` will be set to True. 
                                    If it's None, then `Result.is_positive` is used.
                                    Default is None.
                                    
        Note:
            If `threshold` is set:
            :meth:`Result.is_positive` will be overwritten.
            :meth:`Result.confidence` should be set.
        """
        self.pre_evaluate()

        for trial_result in self._results:
            if threshold is not None:
                trial_result.is_positive = False
                if trial_result.confidence >= threshold:
                    trial_result.is_positive = True

            gt_positive = self._ground_truth.is_positive(trial_result.record1.id, trial_result.record2.id)
            trial_positive = trial_result.is_positive

            if trial_positive and gt_positive:
                self.tp_list.append(trial_result)
            elif not trial_positive and not gt_positive:
                self.tn_list.append(trial_result)
            elif trial_positive and not gt_positive:
                self.fp_list.append(trial_result)
            elif not trial_positive and gt_positive:
                self.fn_list.append(trial_result)

        self.tp = len(self.tp_list)
        self.tn = len(self.tn_list)
        self.fp = len(self.fp_list)
        self.fn = len(self.fn_list)

    def run_munkres(self, threshold=0):
        """
        Run Munkres algorithm (also called the Hungarian algorithm) on all pairs in Trial. 
        Only run this method if the linkage between two datasets are one to one.
        
        Args:
            threshold (float, optional): Only if :meth:`Result.confidence` is greater than this threshold,
                                    `Result.is_positive` will be set to True.
                                    Default is 0.
        
        Note:
            :meth:`Result.is_positive` will be overwritten.
            :meth:`Result.confidence` should be set.
        """
        r1ids = [r.record1.id for r in self._results]
        r2ids = [r.record2.id for r in self._results]
        confs = [r.confidence for r in self._results]
        r1_idx = {v: i for i, v in enumerate(set(r1ids))}  # id -> index
        r2_idx = {v: i for i, v in enumerate(set(r2ids))}  # id -> index
        matrix = [len(r2_idx) * [1] for _ in range(len(r1_idx))]
        for i in range(len(r1ids)):
            matrix[r1_idx[r1ids[i]]][r2_idx[r2ids[i]]] = 1.0 - confs[i]

        # TODO:
        # replace Munkres here by an implementation supports sparse matrix

        row_idx, col_idx = linear_sum_assignment(matrix)
        indexes = set([(r, c) for r, c in zip(row_idx, col_idx)])

        for trial_result in self._results:
            trial_result.is_positive = False
            if (r1_idx[trial_result.record1.id], r2_idx[trial_result.record2.id]) in indexes:
                if trial_result.confidence >= threshold:
                    trial_result.is_positive = True

    def add_result(self, record1: Record, record2: Record, is_positive: bool, confidence: float = 1, **kwargs) -> None:
        """
        Add comparison result

        Args:
            record1 (Record): first record.
            record2 (Record): second record.
            is_positive (bool): if the prediction of these two records is pair
            confidence (float): the probability of positive
        """
        if confidence >= self._min_confidence and self._ground_truth.is_member(record1.id, record2.id):
            if self._top_k == 0 or len(self._results) < self._top_k:
                cur = self.Result(record1, record2, is_positive, confidence, **kwargs)
                heapq.heappush(self._results, cur)
            elif confidence > self._results[0].confidence:
                heapq.heappop(self._results)
                cur = self.Result(record1, record2, is_positive, confidence, **kwargs)
                heapq.heappush(self._results, cur)

    def add_positive(self, *args, **kwargs):
        """
        Syntactic sugar of :meth:`add_result`
        """
        self.add_result(*args, is_positive=True, **kwargs)

    def add_negative(self, *args, **kwargs):
        """
        Syntactic sugar of :meth:`add_result`
        """
        self.add_result(*args, is_positive=False, **kwargs)

    def get_all_data(self):
        """
        Get all saved Result

        Returns:
            list[Result]:
        """
        return self._results

    def get_ground_truth(self):
        """
        Get associated GroundTruth

        Returns:
            GroundTruth:
        """
        return self._ground_truth

    @property
    def precision(self) -> float:
        """
        precision = true positive / (true positive + false positive)

        Returns:
            float:
        """
        if (self.tp + self.fp) == 0:
            return 0.0
        return self.tp / (self.tp + self.fp)

    @property
    def recall(self) -> float:
        """
        recall = true positive / (true positive + false negative)

        Returns:
            float:
        """
        if (self.tp + self.fn) == 0:
            return 0.0
        return self.tp / (self.tp + self.fn)

    @property
    def f_measure(self) -> float:
        """
        f_measure = 2 * precision * recall / (precision + recall)

        Returns:
            float:
        """
        return 2 * (self.precision * self.recall) / (self.precision + self.recall)
        
    @property
    def false_positives(self) -> float:
        """
        false positive ratio = false positive / (false positive + true negative)

        Returns:
            float:
        """
        if (self.fp + self.tn) == 0:
            return 0.0
        return self.fp / (self.fp + self.tn)

    @property
    def true_positives(self) -> float:
        """
        true positive ratio = true positive / (true positive + false negative)

        Returns:
            float:
        """
        if (self.tp + self.fn) == 0:
            return 0.0
        return self.tp / (self.tp + self.fn)

    @property
    def false_negatives(self) -> float:
        """
        false negative ratio = false negative / (false negative + true positive)

        Returns:
            float:
        """
        if (self.tp + self.fn) == 0:
            return 0.0
        return self.fn / (self.tp + self.fn)

    @property
    def true_negatives(self) -> float:
        """
        true negative ratio = true negative / (true negative + false positive)

        Returns:
            float:
        """
        if (self.fp + self.tn) == 0:
            return 0.0
        return self.tn / (self.fp + self.tn)

    @property
    def false_discovery(self):
        """
        false discovery = false positive / (false positive + true positive)

        Returns:
            float:
        """
        if (self.fp + self.tp) == 0:
            return 0.0
        return self.fp / (self.fp + self.tp)

    @property
    def true_positives_list(self):
        """
        List of all true positives
        
        Returns:
            list:
        """
        return self.tp_list

    @property
    def true_negatives_list(self):
        """
        List of all true negatives
        
        Returns:
            list:
        """
        return self.tn_list

    @property
    def false_positives_list(self):
        """
        List of all false positives
        
        Returns:
            list:
        """
        return self.fp_list

    @property
    def false_negatives_list(self):
        """
        List of all false negatives
        
        Returns:
            list:
        """
        return self.fn_list

    def generate_dataframe(self, results, record1_columns=None, record2_columns=None, result_columns=None, **kwargs):
        """
        Generate Pandas Dataframe
        
        Args:
            results (list): Result list
            record1_columns (list, optional): List of property names from record 1 which need to be shown in dataframe columns.
                                            Default is None, all properties are used.
            record2_columns (list, optional): List of property names from record 2 which need to be shown in dataframe columns.
                                            Default is None, all properties are used.
            result_columns (list, optional): List of property names from result which need to be shown in dataframe columns.
                                            Default is None, all properties are used.
            **kwargs: Parameters of pandas.Dataframe.
            
        Returns:
            pandas.DataFrame:
        """

        table = []
        r1_columns = record1_columns
        r2_columns = record2_columns
        res_columns = result_columns

        # construct table
        for result in results:

            # generate columns based on first result
            if not r1_columns:
                r1_columns = get_property_names(result.record1.__class__)
            if not r2_columns:
                r2_columns = get_property_names(result.record2.__class__)
            if not res_columns:
                res_columns = result.get_property_names()

            # get data
            r1_data = []
            r2_data = []
            res_data = []
            gt_data = []
            for prop_name in r1_columns:
                r1_data.append(getattr(result.record1, prop_name))
            for prop_name in r2_columns:
                r2_data.append(getattr(result.record2, prop_name))
            for prop_name in res_columns:
                res_data.append(getattr(result, prop_name))
            gt_data.append(self._ground_truth.get_label(result.record1.id, result.record2.id))

            # append data
            table.append(r1_data + r2_data + gt_data + res_data)

        r1_columns = ['record1.{}'.format(p) for p in r1_columns]
        r2_columns = ['record2.{}'.format(p) for p in r2_columns]
        columns = r1_columns + r2_columns + ['ground_truth.label'] + res_columns
        return pd.DataFrame(table, columns=columns, **kwargs)



import collections
import math

import rltk.utils as utils


def tf_idf_similarity(bag1, bag2, df_corpus, doc_size, math_log=False):
    """
    Computes TF/IDF measure. This measure employs the notion of TF/IDF score commonly used in information retrieval (IR) to find documents that are relevant to keyword queries. The intuition underlying the TF/IDF measure is that two strings are similar if they share distinguishing terms.
        
    Note:
        If you will call this function many times, :meth:`TF_IDF` is more efficient.

    Args:
        bag1 (list): Bag 1.
        bag2 (list): Bag 2.
        df_corpus (dict): The pre calculated document frequency of corpus.
        doc_size (int): total documents used in corpus.
        math_log (bool, optional): Flag to indicate whether math.log() should be used in TF and IDF formulas. Defaults to False.

    Returns:
        float: TF/IDF cosine similarity.

    Examples:
        >>> rltk.tfidf(['a', 'b', 'a'], ['a', 'c'], {'a':3, 'b':1, 'c':1}, 3)
        0.17541160386140586
        >>> rltk.tfidf(['a', 'b', 'a'], ['a', 'c'], {'a':3, 'b':2, 'c':1}, 4, True)
        0.12977804138
        >>> rltk.tfidf(['a', 'b', 'a'], ['a'], {'a':3, 'b':1, 'c':1}, 3)
        0.5547001962252291
    """
    # http://www.tfidf.com/

    utils.check_for_none(bag1, bag2, df_corpus)
    utils.check_for_type(list, bag1, bag2)

    # term frequency for input strings
    t_x, t_y = collections.Counter(bag1), collections.Counter(bag2)
    tf_x = {k: float(v) / len(bag1) for k, v in t_x.items()}
    tf_y = {k: float(v) / len(bag2) for k, v in t_y.items()}

    # unique element
    total_unique_elements = set()
    total_unique_elements.update(bag1)
    total_unique_elements.update(bag2)

    idf_element, v_x, v_y, v_x_y, v_x_2, v_y_2 = 0.0, 0.0, 0.0, 0.0, 0.0, 0.0

    # tfidf calculation
    for element in total_unique_elements:
        if element not in df_corpus:
            continue
        idf_element = doc_size * 1.0 / df_corpus[element]

        v_x = 0 if element not in tf_x else (math.log(idf_element) * tf_x[element]) if math_log else (
            idf_element * tf_x[element])
        v_y = 0 if element not in tf_y else (math.log(idf_element) * tf_y[element]) if math_log else (
            idf_element * tf_y[element])
        v_x_y += v_x * v_y
        v_x_2 += v_x * v_x
        v_y_2 += v_y * v_y

    # cosine similarity
    return 0.0 if v_x_y == 0 else v_x_y / (math.sqrt(v_x_2) * math.sqrt(v_y_2))


def compute_tf(tokens):
    """
    Compute TF (Term Frequency)
    
    Args:
        tokens (list): tokens
    """
    terms = collections.Counter(tokens)
    return {k: float(v) / len(tokens) for k, v in terms.items()}


def compute_idf(df_corpus, doc_size, math_log=False):
    """
    Compute IDF (Inverted Document Frequency)
    
    Args:
        df_corpus (dict): terms in document
        doc_size (int): total document size
        math_log (bool): logarithm of the result
    """
    return {k: doc_size * 1.0 / v if math_log is False \
        else math.log(doc_size * 1.0 / v) \
            for k, v in df_corpus.items()}


def tf_idf_cosine_similarity(tfidf_dict1, tfidf_dict2):
    """
    Compute Cosine similarity for TF/IDF value dictionary
    
    Args:
        tfidf_dict1 (dict): TF/IDF dictionary for first record, format in ``{term1: tfidf value, ...}``
        tfidf_dict2 (dict): TF/IDF dictionary for second record, same format as tfidf_dict1.
    
    Returns:
        float:
    """
    v_x_y, v_x_2, v_y_2 = 0.0, 0.0, 0.0

    # intersection of dict1 and dict2
    # ignore the values that are not in both
    for t in tfidf_dict1.keys():
        if t in tfidf_dict2:
            v_x_y += tfidf_dict1[t] * tfidf_dict2[t]

    for t, tfidf in tfidf_dict1.items():
        v_x_2 += tfidf * tfidf
    for t, tfidf in tfidf_dict2.items():
        v_y_2 += tfidf * tfidf

    # cosine similarity
    return 0.0 if v_x_y == 0 else v_x_y / (math.sqrt(v_x_2) * math.sqrt(v_y_2))


class TF_IDF():
    """
    TF/IDF helper class (An efficient implementation)
    
    Examples::
    
        # initialization
        tfidf = TF_IDF()
        # add document
        tfidf.add_document('id1', ['a', 'b', 'a'])
        tfidf.add_document('id2', ['b', 'c'])
        tfidf.add_document('id3', ['b', 'd'])
        # compute idf
        tfidf.pre_compute()
        # get similarity
        tfidf.similarity('id1', 'id2')
        tfidf.similarity('id1', 'id3')
    """

    def __init__(self):
        self.tf = {}
        self.df_corpus = {}
        self.doc_size = 0
        self.idf = 0

    def add_document(self, doc_id: str, tokens: list):
        """
        Add document to corpus
        
        Args:
            doc_id (str): Document (record) id.
            tokens (list): List of token string.
        """
        self.doc_size += 1
        tf = compute_tf(tokens)
        self.tf[doc_id] = tf
        for k, _ in tf.items():
            self.df_corpus[k] = self.df_corpus.get(k, 0) + 1

    def pre_compute(self, math_log: bool = False):
        """
        Pre-compute IDF score
        
        Args:
            math_log (bool, optional): Flag to indicate whether math.log() should be used in TF and IDF formulas. Defaults to False.
        """
        self.idf = compute_idf(self.df_corpus, self.doc_size, math_log)

    def similarity(self, id1, id2):
        """
        Get similarity
        
        Args:
            id1 (str): id 1
            id2 (str): id2
        
        Returns:
            float:
        """
        tf_x = self.tf[id1]
        tfidf_x = {k: v * self.idf[k] for k, v in tf_x.items()}
        tf_y = self.tf[id2]
        tfidf_y = {k: v * self.idf[k] for k, v in tf_y.items()}
        return tf_idf_cosine_similarity(tfidf_x, tfidf_y)



# https://docs.scipy.org/doc/scipy-0.14.0/reference/spatial.distance.html
from scipy.spatial.distance import euclidean, cityblock

import rltk.utils as utils


def euclidean_distance(vec1, vec2, weights=None):
    """
    Euclidean distance.

    Args:
        vec1 (list): Vector 1. List of integer or float.
        vec2 (list): Vector 2. List of integer or float. It should have the same length to vec1.
        weights (list): Weights for each value in vectors. If it's None, all weights will be 1.0. Defaults to None.

    Returns:
        float: Euclidean distance.
    """

    utils.check_for_none(vec1, vec2)
    utils.check_for_type(list, vec1, vec2)
    if weights:
        utils.check_for_type(list, weights)
    if len(vec1) != len(vec2):
        raise ValueError('vec1 and vec2 should have same length')

    return euclidean(vec1, vec2, weights)


def manhattan_distance(vec1, vec2, weights=None):
    """
    Manhattan distance.

    Args:
        vec1 (list): Vector 1. List of integer or float.
        vec2 (list): Vector 2. List of integer or float. It should have the same length to vec1.
        weights (list): Weights for each value in vectors. If it's None, all weights will be 1.0. Defaults to None.

    Returns:
        float: Manhattan distance.
    """
    utils.check_for_none(vec1, vec2)
    utils.check_for_type(list, vec1, vec2)
    if weights:
        utils.check_for_type(list, weights)
    if len(vec1) != len(vec2):
        raise ValueError('vec1 and vec2 should have same length')

    return cityblock(vec1, vec2, weights)



import json


from rltk.io.reader import Reader
from rltk.io.io_utils import get_file_handler


class JsonLinesReader(Reader):
    """
    `JSON Lines <http://jsonlines.org/>`_ Reader.
    
    Args:
        file_handler (str/io.IOBase): File name or file handler of input file.
        ignore_blank_line (bool): If blank line should be ignored. Defaults to True.
    """

    def __init__(self, file_handler, ignore_blank_line=True):
        self._file_handler = get_file_handler(file_handler)
        self._ignore_blank_line = ignore_blank_line

    def __next__(self):
        for line in self._file_handler:
            if line.strip() == '':
                if self._ignore_blank_line:
                    continue
                else:
                    raise ValueError('Blank line detected')
            yield json.loads(line)

    def close(self):
        try:
            self._file_handler.close()
        except:
            pass



import rltk.utils as utils


def _lcs(s1, s2):
    m, n = len(s1), len(s2)

    dp = [[None] * (n + 1) for i in range(m + 1)]

    for i in range(m + 1):
        for j in range(n + 1):
            if i == 0 or j == 0:
                dp[i][j] = 0
            elif s1[i - 1] == s2[j - 1]:
                dp[i][j] = dp[i - 1][j - 1] + 1
            else:
                dp[i][j] = max(dp[i - 1][j], dp[i][j - 1])

    return dp[m][n]


def longest_common_subsequence_distance(s1, s2):
    """
    The LCS distance between strings X (of length n) and Y (of length m) is n + m - 2 |LCS(X, Y)| min = 0 max = n + m

    Args:
        s1 (str): Sequence 1.
        s2 (str): Sequence 2.

    Returns:
        float: Longest Common Subsequence Distance.

    Examples:
        >>> rltk.longest_common_subsequence_distance('abcd', 'acbd')
        2
        >>> rltk.longest_common_subsequence_distance('abcdefg', 'acef')
        3
    """
    utils.check_for_none(s1, s2)
    utils.check_for_type(str, s1, s2)

    m, n = len(s1), len(s2)

    # dp = [[None] * (n + 1) for i in range(m + 1)]

    lcs = _lcs(s1, s2)
    return n + m - 2 * lcs


def metric_longest_common_subsequence(s1, s2):
    """
    The Metric LCS distance between 2 strings is similar to LCS between 2 string where Metric Longest Common Subsequence is computed as 1 - |LCS(s1, s2)| / max(|s1|, |s2|)

    Args:
        s1 (str): Sequence 1.
        s2 (str): Sequence 2.

    Returns:
        float: Metric Longest Common Subsequence Distance.

    Examples:
        >>> rltk.longest_common_subsequence('ABCDEFG', 'ABCDEFHJKL')
        0.4
        # LCS: ABCDEF => length = 6
        # longest = s2 => length = 10
        # => 1 - 6/10 = 0.4

        >>> rltk.optimal_string_alignment_distance('ABDEF', 'ABDIF')
        4
        # LCS: ABDF => length = 4
        # longest = ABDEF => length = 5
        # => 1 - 4 / 5 = 0.2
    """
    utils.check_for_none(s1, s2)
    utils.check_for_type(str, s1, s2)

    lcs = _lcs(s1, s2)
    return 1 - float(lcs) / max(len(s1), len(s2), 1)
